from typing import Optional

from pypaq.R4C.policy_gradients.base.pt_based.pg_PT_actor import PG_PTActor
from pypaq.R4C.policy_gradients.a2c.pt_based.a2c_PT_module import A2CModel
from pypaq.torchness.motorch import Module



class A2C_PTActor(PG_PTActor):

    def __init__(
            self,
            name: str=                          'A2C_PTActor',
            nngraph: Optional[type(Module)]=    A2CModel,
            **kwargs):
        PG_PTActor.__init__(
            self,
            name=       name,
            nngraph=    nngraph,
            **kwargs)


"""

 2019 (c) piteren

"""

import tensorflow as tf
from tensorflow.contrib import rnn

from pypaq.neuralmess.base_elements import my_initializer
from pypaq.neuralmess.layers import lay_dense


# N prediction decoder
def decN(
        input,
        dictW,
        predN=          1,          # N samples for every feature
        name=           'decN',
        hLays=          None,       # tuple or list of ints
        hActiv=         tf.nn.relu,
        initializer=    None,
        seed=           12321,
        verbLev=        0):

    if verbLev > 0: print('\nBuilding decoderN ...')
    if verbLev > 1: print('decoder input:', input)

    if initializer is None: initializer = my_initializer(seed)

    with tf.variable_scope(name):

        # hidden layers
        if hLays:
            for nLay in range(len(hLays)):
                laySize = hLays[nLay]
                input = lay_dense(
                    input=          input,
                    units=          laySize,
                    activation=     hActiv,
                    use_bias=       True,
                    initializer=    initializer,
                    seed=           seed,
                    name=           'decoderN_Hlay_%s' % nLay)

        # projection to predN x dictW
        logits = lay_dense(
            input=          input,
            units=          predN * dictW,
            activation=     None,
            use_bias=       True,
            initializer=    initializer,
            seed=           seed,
            name=           'decoderNProjection')
        if verbLev > 1: print(' > projection to logits (%dx dictW):' % predN, logits)

        if predN > 1:
            logits = tf.reshape(logits, [tf.shape(logits)[0], -1, dictW])
            if verbLev > 1: print(' > reshaped logits (B,%dxS,dictW):' % predN, logits)

        predictions = tf.argmax(logits, axis=-1, output_type=tf.int32)
        if verbLev > 1: print(' > predictions:', predictions)

    return logits, predictions

# attention rnn decoder
def decARNN(
        input,
        target,
        embWeights,
        name=           'decARNN',
        cellFN=         rnn.NASCell,
        cellWidth=      64,
        attentionType=  'Luong',
        attMechSize=    72,             # att mechanism layer size (for Luong == cellWidth)
        attLaySize=     52,             # (opt) IN: concat(rnn+context), no bias => projectionL & rnnIN
        verbLev=        0):

    with tf.variable_scope(name):
        suppAttTypes = ['Luong', 'Bahdanau']
        assert attentionType in suppAttTypes, '>>> Err: unsupported attention type'
        attType = tf.contrib.seq2seq.BahdanauAttention
        if attentionType == 'Luong':
            attType = tf.contrib.seq2seq.LuongAttention
            attMechSize = cellWidth

        if verbLev > 1: print('decoder inputs:', input)
        if verbLev > 1: print('decoder targets:', target)
        targetsShape = target.shape

        # prepare targets with <go> token at the beginning
        goLabel = embWeights.shape[-1] - 1
        goLabelTensor = tf.fill([targetsShape[0], 1], goLabel)
        goTargets = tf.concat([goLabelTensor, target[:, :targetsShape[1] - 1]], -1)
        if verbLev > 1: print('decoder targets with go:', goTargets)

        # targets embedding for helper (decoder cell input)
        embTargets = tf.nn.embedding_lookup(params=embWeights, ids=goTargets)

        decoderCell = cellFN(cellWidth)
        attMechanism = attType(num_units=attMechSize,
                               memory=input)
        decoderCell = tf.contrib.seq2seq.AttentionWrapper(
            cell=                   decoderCell,
            attention_mechanism=    attMechanism,
            attention_layer_size=   attLaySize)

        tSEQlen = tf.fill(dims=[targetsShape[0]], value=targetsShape[1])
        if verbLev > 1: print('decoder tSEQlen:', tSEQlen)
        helper = tf.contrib.seq2seq.TrainingHelper(
            inputs=         embTargets,
            sequence_length=tSEQlen)
        initialState = decoderCell.zero_state(targetsShape[0], tf.float32)
        projection_layer = tf.layers.Dense(
            units=          embWeights.shape[0],
            use_bias=        False,
            name=           'decoderProjection')
        decoder = tf.contrib.seq2seq.BasicDecoder(
            cell=           decoderCell,
            helper=         helper,
            initial_state=  initialState,
            output_layer=   projection_layer)

        output, _, _ = tf.contrib.seq2seq.dynamic_decode(decoder=decoder)
        if verbLev > 1: print('decoder output:', output)

        logits = output.rnn_output
        labels = output.sample_id
    return logits, labels


"""

 2018 (c) piteren

"""

import logging
import os
import sys
from typing import Optional

from pypaq.lipytools.little_methods import stamp
from pypaq.lipytools.files import prep_folder


# logger duplicates print() output to given file
class Logger:

    def __init__(
            self,
            file_name):
        self.terminal = sys.stdout
        self.log = open(file_name, 'a', encoding='utf-8')

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)
        self.log.flush()

    def flush(self):
        self.log.flush()

    # only to pass unittests with objects that uses logger
    def getvalue(self): return 0


class Logg:

    ERROR =     logging.ERROR   # 0
    WARNING =   logging.WARNING # 0
    INFO =      logging.INFO    # 1
    DEBUG =     logging.DEBUG   # 2

    def __init__(
            self,
            name=                   'logger',
            folder: Optional[str]=  None,
            level=                  None):
        if not level: level = logging.INFO
        filename = f'{folder}/{name}.log' if folder else None
        self.logger = logging
        self.logger.basicConfig(
            level=      level,
            format=     '%(message)s',
            filename=   filename)
        fni = f', file: {filename}' if filename else ''
        self(f'Logger {name} started, level: {self.logger.getLevelName(level=level)}{fni}')


    def __call__(self, msg:str, level=None):
        if level == Logg.DEBUG:             self.logger.debug(msg)
        if level == Logg.INFO or not level: self.logger.info(msg)
        if level == Logg.WARNING:           self.logger.warning(f'WARNING: {msg}')
        if level == Logg.ERROR:             self.logger.error(f'ERROR: {msg}')


# method setting logger to logfile, returns path to logfile
def set_logger(
        log_folder: str,
        custom_name: Optional[str]= None,
        add_stamp=                  True,
        verb=                       1) -> str:

    prep_folder(log_folder)

    file_name = custom_name if custom_name else 'run'
    if add_stamp: file_name += f'_{stamp()}'
    file_name += '.log'

    logfile_path = os.path.join(log_folder, file_name)
    sys.stdout = Logger(logfile_path)

    if verb>0:
        script_path = os.path.join(os.path.dirname(sys.argv[0]), os.path.basename(sys.argv[0]))
        print(f'\nLogger started..')
        print(f' > for script : {script_path}')
        print(f' > logfile    : {logfile_path}')

    return logfile_path


"""

 2022 (c) piteren

    QTableTrainer - Trainer specific for training of QTableActor, sets update_rate of Actor.

"""

from pypaq.R4C.qlearning.ql_trainer import QLearningTrainer
from pypaq.R4C.qlearning.qtable.qt_actor import QTableActor


class QTableTrainer(QLearningTrainer):

    def __init__(
            self,
            actor: QTableActor,
            update_rate: float,
            **kwargs):

        QLearningTrainer.__init__(self, actor=actor, **kwargs)
        self.actor = actor
        self.actor.set_update_rate(update_rate)

        self._rlog.info('*** QTableTrainer *** initialized')
        self._rlog.info('> actor update_rate: {update_rate}')


"""

 2022 (c) piteren

 QLearningActor
    - supports only finite actions space environments (FiniteActionsRLEnvy)
    - adds QLearning methods

"""

from abc import abstractmethod, ABC
import numpy as np
from typing import List

from pypaq.R4C.actor import TrainableActor
from pypaq.R4C.envy import FiniteActionsRLEnvy
from pypaq.lipytools.softmax import softmax


class QLearningActor(TrainableActor, ABC):

    def __init__(
            self,
            envy: FiniteActionsRLEnvy,
            seed: int,
            **kwargs):

        TrainableActor.__init__(self, envy=envy, **kwargs)
        self._envy = envy  # to update type (for pycharm only)

        np.random.seed(seed)

        self._rlog.info('*** QLearningActor *** initialized')
        self._rlog.info(f'> num_actions: {self._envy.num_actions()}')
        self._rlog.info(f'> seed:        {seed}')

    # returns QVs (QV for all actions) for given observation
    @abstractmethod
    def _get_QVs(self, observation: object) -> np.ndarray: pass

    # returns QVs (for all actions) for given observations batch, here baseline implementation - may be overridden with optimized version
    def get_QVs_batch(self, observations: List[object]) -> np.ndarray:
        return np.array([self._get_QVs(o) for o in observations])

    def get_policy_action(
            self,
            observation: object,
            sampled=    False) -> int:

        qvs = self._get_QVs(observation)

        if sampled:
            obs_probs = softmax(qvs) # baseline with softmax on QVs
            return np.random.choice(len(qvs), p=obs_probs)

        return int(np.argmax(qvs))

    # updates QV for given observation and action (single observation/action), returns Actor "metric" - loss etc. (float)
    @abstractmethod
    def _upd_QV(
            self,
            observation: object,
            action: int,
            new_qv: float) -> float: pass

    # updates QV for given observations and actions batch, may be overridden with optimized version
    def update_with_experience(
            self,
            observations: List[object], # batch of observations
            actions: List[int],         # batch of selected actions for given observations (do not have to come from Actor policy!)
            new_qvs: List[float],       # batch of QV to be updated (QV for selected action only)
            inspect=    False) -> dict:
        loss = 0.0
        for ob, ac, nq in zip(observations, actions, new_qvs):
            loss += self._upd_QV(
                observation=    ob,
                action=         ac,
                new_qv=         nq)
        return {'loss': loss}


from abc import ABC, abstractmethod
from multiprocessing import cpu_count, Process, Queue
from queue import Empty
import psutil
import time
from typing import Any, Optional

from pypaq.lipytools.pylogger import get_pylogger


# message sent between processes via Ques (my que)
class QMessage:
    def __init__(self, type:str, data:Any):
        self.type = type
        self.data = data

# my que
class Que:

    def __init__(self):
        self.q = Queue()

    def put(self, obj:QMessage, **kwargs):
        assert isinstance(obj, QMessage)
        self.q.put(obj, **kwargs)

    def get(self, **kwargs) -> QMessage:
        obj = self.q.get(**kwargs)
        assert isinstance(obj, QMessage)
        return obj

    # non blocking get
    def get_if(self) -> Optional[QMessage]:
        try:
            obj = self.q.get_nowait()
            assert isinstance(obj, QMessage)
            return obj
        except Empty:
            return None

    def get_timeout(self, timeout:float, **kwargs) -> Optional[QMessage]:
        try:
            obj = self.q.get(timeout=timeout, **kwargs)
            assert isinstance(obj, QMessage)
            return obj
        except Empty:
            return None

    def empty(self): return self.q.empty()

    def qsize(self): return self.q.qsize()


# Exception Managed Subprocess - with basic exceptions management
class ExSubprocess(Process, ABC):

    def __init__(
            self,
            ique: Que,                          # input que
            oque: Que,                          # output que
            id: Optional[int or str]=   None,   # unique id to identify the subprocess, if not given takes from Process name
            raise_unk_exception=        True,   # raises exception other than KeyboardInterrupt
            logger=                     None,
            loglevel=                   20):

        ABC.__init__(self)
        Process.__init__(self, target=self.__run)
        if id is None: id = self.name

        self.id = id

        if not logger:
            logger = get_pylogger(
                name=       self.id,
                add_stamp=  False,
                folder=     None,
                level=      loglevel)
        self.logger = logger

        self.raise_unk_exception = raise_unk_exception

        self.ique = ique
        self.oque = oque

        self.logger.info(f'*** ExSubprocess *** id: {self.id} initialized')

    # process target method, wraps subprocess_method() with try / except
    def __run(self):
        try:
            self.logger.info(f'> ExSubprocess ({self.id}, pid:{self.pid}) - starting subprocess_method()')
            self.subprocess_method()
            self.logger.info(f'> ExSubprocess ({self.id}, pid:{self.pid}) - finished subprocess_method()')
        except KeyboardInterrupt:
            self.__exception_handle('KeyboardInterrupt')
        except Exception as e:
            self.__exception_handle(f'other: {str(e)}')
            if self.raise_unk_exception: raise e

    # method run in subprocess, to be implemented
    @abstractmethod
    def subprocess_method(self): pass

    # when exception occurs, message with exception data is put on the output que
    def __exception_handle(self, name: str):
        msg = QMessage(
            type=   f'ex_{name}, ExSubprocess id: {self.id}, pid: {self.pid}',
            data=   self.id) # returns ID here to allow process identification
        self.oque.put(msg)
        self.logger.warning(f'> ExSubprocess ({self.id}) halted by exception: {name}')
        self.after_exception_handle_run()

    # this method may be implemented and will be run after exception occurred
    def after_exception_handle_run(self): pass

    def kill(self):
        if self.alive: super().kill()
        while self.alive: time.sleep(0.01)

    def terminate(self):
        if self.alive: super().terminate()
        while self.alive: time.sleep(0.01)

    @property
    def alive(self):
        if self.closed: return False
        return self.is_alive()

    @property
    def closed(self):
        return self._closed

    def get_info(self) -> str:
        pid = self.pid if not self.closed else '<pid:closed>'
        exitcode = self.exitcode if not self.closed else '<exitcode:closed>'
        nfo = f'{str(self)}, pid: {pid}, parent pid: {self._parent_pid}, alive: {self.alive}, exitcode: {exitcode}'
        return nfo


def sys_res_nfo():
    vm = psutil.virtual_memory()
    gb = 1024 ** 3
    return {
        'cpu_count':    cpu_count(),
        'mem_total_GB': vm.total / gb,
        'mem_used_%':   vm.percent}


"""

 2021 (c) piteren

"""

from typing import List

from pypaq.lipytools.plots import histogram


# prepares some stats for list of documents, assumes that each document is a list of sentences, uses whs tokenization for words
def docs_stats(
        docs: List[list],       # list of sentence lists
        print_pd=       True,   # to print stats with pandas
        save_FD :str=   None):  # to save plots to folder


    print(f'\nStats for {len(docs)} documents:')

    vals = {
        'n_sen_doc':    [], # num sentences  per doc
        'n_wrd_doc':    [], # num words      per doc
        'n_chr_doc':    [], # num chars      per doc
        'n_wrd_sen':    [], # num words      per sentence
        'n_chr_sen':    [], # num chars      per sentence
        'n_chr_wrd':    []} # num chars      per word

    for doc in docs:
        doc_words = 0
        doc_chars = 0
        for sen in doc:
            n_chars = len(sen)
            doc_chars += n_chars
            vals['n_chr_sen'].append(n_chars)

            words = sen.split()
            n_words = len(words)
            doc_words += n_words
            vals['n_wrd_sen'].append(n_words)
            for word in words:
                vals['n_chr_wrd'].append(len(word))

        vals['n_sen_doc'].append(len(doc))
        vals['n_wrd_doc'].append(doc_words)
        vals['n_chr_doc'].append(doc_chars)

    for k in vals:
        if print_pd: print(f'\n > {k} stats:')
        histogram(
            val_list=       vals[k],
            name=           k,
            pandas_stats=   print_pd,
            save_FD=        save_FD)


if __name__ == '__main__':

    docA = [
        'To jest przykładowe zdanie.',
        'Inne zdanie, inne od pierwszego.',
        'Kolejne zdanie, które jest w dokumencie.']
    docB = [
        'W dokumencie B są zupełnie inne zdania.',
        'Trudno je nawet zliczyć.',
        'Ale można do tego użyć narzędzi statystycznych.',
        'Mozna to zrobić w python\'ie.']
    docC = [
        'W tym dokumencie jest tylko jedno zdanie, ale za to najdłuższe']

    docs = [docA,docB,docC]

    docs_stats(docs)


"""

 2022 (c) piteren

    QTableActor: QLearningActor with QTable

    QTableActor may be trained by QLearningTrainer. Trainer is responsible for computation of new QV for an Actor.
    Actor on the other side has its own QV update_rate, which works as a kind of Actor-specific learning ratio.
    Learning process is mainly directed by Trainer with possible little override by an Actor.
    Actor is responsible for computation of its loss.

"""

import numpy as np
from typing import Hashable, Dict, List

from pypaq.R4C.qlearning.ql_actor import QLearningActor
from pypaq.R4C.helpers import RLException


class QTable:

    def __init__(self, width: int):
        self.__width = width
        self.__table: Dict[Hashable, np.ndarray] = {}  # {observation_hash: QVs}
        self.__keys: List[np.ndarray] = []

    @staticmethod
    def __hash(observation: np.ndarray) -> str:
        return str(observation)

    def __init_hash(self, ha: str):
        self.__table[ha] = np.zeros(self.__width, dtype=float)

    def get_QVs(self, observation: np.ndarray) -> np.ndarray:
        ha = QTable.__hash(observation)
        if ha not in self.__table:
            self.__init_hash(ha)
            self.__keys.append(observation)
        return self.__table[ha]

    def put_QV(self,
            observation: np.ndarray,
            action: int,
            new_qv: float):
        ha = QTable.__hash(observation)
        if ha not in self.__table:
            self.__init_hash(ha)
            self.__keys.append(observation)
        self.__table[ha][action] = new_qv

    def __str__(self):
        s = f'length: {len(self.__table) if self.__table else "<empty>"}\n'
        if self.__table:
            for k in sorted(self.__table.keys()):
                s += f'{k} : {self.__table[k]}\n'
        return s[:-1]


# implements QLearningActor with QTable
class QTableActor(QLearningActor):

    def __init__(
            self,
            name: str=          'QTableActor',
            save_topdir: str=   '_models',
            **kwargs):

        QLearningActor.__init__(self, name=name, **kwargs)

        self._save_topdir = save_topdir
        self.__qtable = QTable(self._envy.num_actions())
        self._update_rate = None # needs to be set before update

    def set_update_rate(self, update_rate:float):
        self._update_rate = update_rate
        self._rlog.info(f'> QTableActor set update_rate to: {self._update_rate}')

    def _get_QVs(self, observation: object) -> np.ndarray:
        obs_vec = self._get_observation_vec(observation)
        return self.__qtable.get_QVs(obs_vec)

    def _upd_QV(
            self,
            observation: object,
            action: int,
            new_qv: float) -> float:
        if self._update_rate is None: raise RLException('Trainer needs to set update_rate of QLearningActor before training!')
        old_qv = self._get_QVs(observation)[action]
        diff = new_qv - old_qv
        self.__qtable.put_QV(
            observation=    self._get_observation_vec(observation),
            action=         action,
            new_qv=         old_qv + self._update_rate * diff)
        return abs(diff)

    def _get_save_topdir(self) -> str:
        return self._save_topdir

    def save(self):
        raise RLException('not implemented')

    def __str__(self):
        return f'QTableActor, QTable:\n{self.__qtable.__str__()}'


"""

 2022 (c) piteren

    RL Trainer for Actor acting on RLEnvy

        Implements generic RL training procedure, that is valid for some algorithms (QTable, PG, AC).
        May be overridden with custom implementation, returns Dict with some training stats

"""

from abc import abstractmethod, ABC
from collections import deque
import numpy as np
import random
import time
from typing import List, Tuple, Optional

from pypaq.lipytools.pylogger import get_pylogger
from pypaq.lipytools.moving_average import MovAvg
from pypaq.R4C.envy import RLEnvy, FiniteActionsRLEnvy
from pypaq.R4C.actor import TrainableActor
from pypaq.R4C.helpers import RLException
from pypaq.torchness.base_elements import TBwr
from pypaq.comoneural.avg_probs import avg_probs
from pypaq.comoneural.zeroes_processor import ZeroesProcessor


# Trainer Experience Memory (deque od dicts)
class ExperienceMemory:

    def __init__(
            self,
            maxsize: int,
            seed: int):
        self.memory = deque(maxlen=maxsize)
        random.seed(seed)

    def append(self, element:dict):
        self.memory.append(element)

    # returns random sample of non-duplicates from memory
    def sample(self, n) -> List[dict]:
        return random.sample(self.memory, n)

    # returns all elements from memory in original order
    def get_all(self) -> List[dict]:
        return list(self.memory)

    def clear(self):
        self.memory.clear()

    def __len__(self):
        return len(self.memory)


# Reinforcement Learning Trainer for Actor acting on RLEnvy
class RLTrainer(ABC):

    def __init__(
            self,
            envy: RLEnvy,
            actor: TrainableActor,
            batch_size: int,        # Actor update batch data size
            memsize_batches: int,   # ExperienceMemory size (in number of batches)
            exploration: float,     # train exploration factor
            train_sampled: float,   # how often move is sampled (vs argmax) while training
            seed: int=      123,
            logger=         None,
            loglevel=       20,
            hpmser_mode=    False):

        self._rlog = logger or get_pylogger(level=loglevel)
        self._rlog.info(f'*** RLTrainer *** initializes..')
        self._rlog.info(f'> Envy:          {envy.__class__.__name__}')
        self._rlog.info(f'> Actor:         {actor.__class__.__name__}, name: {actor.name}')
        self._rlog.info(f'> batch_size:    {batch_size}')
        self._rlog.info(f'> memory size:   {batch_size*memsize_batches}')
        self._rlog.info(f'> exploration:   {exploration}')
        self._rlog.info(f'> train_sampled: {train_sampled}')
        self._rlog.info(f'> seed:          {seed}')

        self.envy = envy
        self.actor = actor
        self.batch_size = batch_size
        self.memsize = self.batch_size * memsize_batches
        self.exploration = exploration
        self.train_sampled = train_sampled
        self.memory: Optional[ExperienceMemory] = None
        self.seed = seed
        np.random.seed(self.seed)
        self.hpmser_mode = hpmser_mode

        self._tbwr = TBwr(logdir=self.actor.get_save_dir()) if not self.hpmser_mode else None
        self._upd_step = 0 # global Trainer update step

        self._zepro = ZeroesProcessor(
            intervals=  (10,50,100),
            tbwr=       self._tbwr) if not self.hpmser_mode else None

    # updates Actor policy, returns dict with Actor "metrics" - loss etc.
    def _update_actor(self, inspect=False) -> dict:
        batch = self.memory.sample(self.batch_size)
        return self.actor.update_with_experience(
            batch=      batch,
            inspect=    inspect)

    # trainer selects exploring action (with Trainer exploratory policy)
    @abstractmethod
    def _get_exploring_action(self) -> object: pass

    # performs one Actor move (action)
    def _ex_move(
            self,
            exploration=    0.0,    # prob pf exploration
            sampled=        0.0     # prob of sampling (vs argmax)
    ) -> Tuple[object,object,float,bool,bool]:

        pre_action_observation = self.envy.get_observation()

        if np.random.rand() < exploration: action = self._get_exploring_action()
        else:                              action = self.actor.get_policy_action(
                                                        observation=    pre_action_observation,
                                                        sampled=        np.random.rand() < sampled)
        reward, is_terminal, won = self.envy.run(action)

        return pre_action_observation, action, reward, is_terminal, won

    # plays (envy) until N steps performed or terminal state
    def play(
            self,
            reset: bool,            # for True starts play from the initial state
            steps: int,
            break_terminal: bool,   # for True breaks play at terminal state
            exploration: float,
            sampled: float,
            render: bool) -> Tuple[List[object], List[object], List[float], List[bool], List[bool]]:

        self._rlog.log(5,f'playing for {steps} steps..')

        if reset: self.envy.reset()

        observations = []
        actions = []
        rewards = []
        terminals = []
        wons = []

        while len(actions) < steps:

            observation, action, reward, is_terminal, won = self._ex_move(
                exploration=    exploration,
                sampled=        sampled)

            observations.append(observation)
            actions.append(action)
            rewards.append(reward)
            terminals.append(is_terminal)
            wons.append(won)

            if render: self.envy.render()

            if is_terminal:
                if break_terminal: break
                self.envy.reset()

        self._rlog.log(5,f'played {len(actions)} steps (break_terminal is {break_terminal})')
        return observations, actions, rewards, terminals, wons

    # plays one episode from reset till terminal state or max_steps
    def _play_episode(
            self,
            exploration: float,
            sampled: float,
            render: bool,
            max_steps: Optional[int]=   None,  # if max steps is given then single play for max_steps is considered to be won
    ) -> Tuple[List[object], List[object], List[float], bool]:

        if max_steps is None and self.envy.get_max_steps() is None:
            raise RLException('Cannot play episode for Envy where max_steps is None and given max_steps is also None')

        observations, actions, rewards, terminals, wons = self.play(
            steps=          max_steps or self.envy.get_max_steps(),
            reset=          True,
            break_terminal= True,
            exploration=    exploration,
            sampled=        sampled,
            render=         render)

        return observations, actions, rewards, wons[-1]

    # plays n episodes, returns (won_factor, avg/reward)
    def test_on_episodes(
            self,
            n_episodes=                 100,
            max_steps: Optional[int]=   None,
    ) -> Tuple[float, float]:
        n_won = 0
        sum_rewards = 0
        for e in range(n_episodes):
            observations, actions, rewards, won = self._play_episode(
                exploration=    0.0,
                sampled=        0.0,
                render=         False,
                max_steps=      max_steps)
            n_won += int(won)
            sum_rewards += sum(rewards)
        return n_won/n_episodes, sum_rewards/n_episodes

    # generic RL training procedure
    def train(
            self,
            num_updates: int,                       # number of training updates
            upd_on_episode=                 False,  # updates on episode finish / terminal (does not wait till batch)
            test_freq=                      100,    # number of updates between test
            test_episodes: int=             100,    # number of testing episodes
            test_max_steps: Optional[int]=  None,   # max number of episode steps while testing
            test_render: bool=              False,  # renders one episode while test
            inspect: bool=                  False,  # for debug / research
            break_ntests: Optional[int]=    None,   # breaks training after all test episodes succeeded N times in a row
    ) -> dict:

        stime = time.time()
        self._rlog.info(f'Starting train for {num_updates} updates..')

        self.memory = ExperienceMemory(
            maxsize=    self.memsize,
            seed=       self.seed)
        self._rlog.info(f'> initialized ExperienceMemory of maxsize {self.memsize}')

        self.envy.reset()
        loss_mavg = MovAvg()
        lossL = []
        n_actions = 0               # total number of train actions
        n_terminals = 0             # number of terminal states reached while training
        last_terminals = 0          # previous number of terminal states
        n_won = 0                   # number of wins while training
        succeeded_row_curr = 0      # current number of succeeded tests in a row
        succeeded_row_max = 0       # max number of succeeded tests in a row
        for uix in range(num_updates):

            # get a batch of data
            new_actions = 0
            while new_actions < self.batch_size:

                observations, actions, rewards, terminals, wons = self.play(
                    steps=          self.batch_size - new_actions,
                    reset=          False,
                    break_terminal= True,
                    exploration=    self.exploration,
                    sampled=        self.train_sampled,
                    render=         False)

                new_actions += len(actions)
                n_actions += len(actions)

                next_observations = observations[1:] + [self.envy.get_observation()]

                # INFO: not all algorithms (QLearning,PG,AC) need all the data below (we store 'more' just in case)
                for o,a,r,n,t in zip(observations, actions, rewards, next_observations, terminals):
                    self.memory.append(dict(observation=o, action=a, reward=r, next_observation=n, terminal=t))

                if terminals[-1]: n_terminals += 1 # ..may not be terminal when limit of new_actions reached
                if wons[-1]: n_won += 1

                self._rlog.debug(f' >> Trainer gots {len(observations):3} observations after play and {len(self.memory):3} in memory, new_actions: {new_actions}' )

                if upd_on_episode: break

            # update Actor & process metrics
            upd_metrics = self._update_actor(inspect=inspect and uix % test_freq == 0)
            self._upd_step += 1

            if 'loss' in upd_metrics: lossL.append(loss_mavg.upd(upd_metrics['loss']))

            # process / monitor policy probs
            if self._tbwr and 'probs' in upd_metrics:
                for k,v in avg_probs(upd_metrics.pop('probs')).items():
                    self._tbwr.add(value=v, tag=f'actor_upd/{k}', step=self._upd_step)

            if self._zepro and 'zeroes' in upd_metrics:
                self._zepro.process(zs=upd_metrics.pop('zeroes'))

            if self._tbwr:
                for k,v in upd_metrics.items():
                    self._tbwr.add(value=v, tag=f'actor_upd/{k}', step=self._upd_step)

            # test Actor
            if uix % test_freq == 0:

                # single episode
                observations, actions, rewards, won = self._play_episode(
                    exploration=    0.0,
                    sampled=        0.0,
                    render=         test_render,
                    max_steps=      test_max_steps)

                # few tests
                avg_won, avg_return = self.test_on_episodes(
                    n_episodes=     test_episodes,
                    max_steps=      test_max_steps)

                self._rlog.info(f'# {uix:3} term:{n_terminals}(+{n_terminals-last_terminals}) -- TS: {len(actions)} actions, return {sum(rewards):.1f} ({"won" if won else "lost"}) -- {test_episodes}xTS: avg_won: {avg_won*100:.1f}%, avg_return: {avg_return:.1f} -- loss_actor: {loss_mavg():.4f}')
                last_terminals = n_terminals

                if avg_won == 1:
                    succeeded_row_curr += 1
                    if succeeded_row_curr > succeeded_row_max: succeeded_row_max = succeeded_row_curr
                else: succeeded_row_curr = 0

            if break_ntests is not None and succeeded_row_curr==break_ntests: break

        self._rlog.info(f'### Training finished, time taken: {time.time()-stime:.2f}sec')

        return { # training_report
            'n_actions':            n_actions,
            'lossL':                lossL,
            'n_terminals':          n_terminals,
            'n_won':                n_won,
            'succeeded_row_max':    succeeded_row_max}


# FiniteActions RL Trainer (for Actor acting on FiniteActionsRLEnvy)
class FATrainer(RLTrainer, ABC):

    def __init__(self, envy:FiniteActionsRLEnvy, seed:int, **kwargs):
        np.random.seed(seed)
        RLTrainer.__init__(self, envy=envy, seed=seed, **kwargs)
        self.envy = envy  # INFO: type "upgrade" for pycharm editor

        self._rlog.info(f'*** FATrainer *** initialized')
        self._rlog.info(f'> number of actions: {self.envy.num_actions()}')

    # selects 100% random action from action space, (np. seed is fixed at Trainer)
    def _get_exploring_action(self):
        return np.random.choice(self.envy.num_actions())


"""

 2022 (c) piteren

    Policy Gradients Trainer

"""
import numpy as np

from pypaq.lipytools.plots import two_dim_multi
from pypaq.R4C.helpers import zscore_norm, extract_from_batch, discounted_return, movavg_return
from pypaq.R4C.policy_gradients.pg_actor import PGActor
from pypaq.R4C.trainer import FATrainer


class PGTrainer(FATrainer):

    def __init__(
            self,
            actor: PGActor,
            discount: float,    # for discounted returns
            use_mavg: bool,     # use movavg to calculate discounted returns
            mavg_factor: float,
            do_zscore: bool,
            **kwargs):

        FATrainer.__init__(self, actor=actor, **kwargs)
        self.actor = actor # INFO: just type "upgrade" for pycharm editor
        self.discount = discount
        self.use_mavg = use_mavg
        self.movavg_factor = mavg_factor
        self.do_zscore = do_zscore

        self._rlog.info('*** PGTrainer *** initialized')
        self._rlog.info(f'> discount: {self.discount}')

    # PGActor update method
    def _update_actor(self, inspect=False) -> dict:

        # get all and flush
        batch = self.memory.get_all()
        self.memory.clear()

        observations = extract_from_batch(batch, 'observation')
        actions =      extract_from_batch(batch, 'action')
        rewards =      extract_from_batch(batch, 'reward')
        terminals =    extract_from_batch(batch, 'terminal')

        # ********************************************************************************************* prepare dreturns

        # split rewards into episodes
        episode_rewards = []
        cep = []
        for r,t in zip(rewards,terminals):
            cep.append(r)
            if t:
                episode_rewards.append(cep)
                cep = []
        if cep: episode_rewards.append(cep)

        if inspect:

            obs_arr = np.array(observations)
            oL = np.split(obs_arr, obs_arr.shape[-1], axis=-1)

            two_dim_multi(
                ys=     oL,
                names=  [f'obs_{ix}' for ix in range(len(oL))])

            # prepare all 4 for inspect
            ret_mavg = []
            ret_disc = []
            for rs in episode_rewards:
                ret_mavg += movavg_return(rewards=rs, factor=self.movavg_factor)
                ret_disc += discounted_return(rewards=rs, discount=self.discount)
            ret_mavg_norm = zscore_norm(ret_mavg)
            ret_disc_norm = zscore_norm(ret_disc)

            two_dim_multi(
                ys=     [
                    rewards,
                    ret_mavg,
                    ret_disc,
                    ret_mavg_norm,
                    ret_disc_norm],
                names=  [
                    'rewards',
                    'ret_mavg',
                    'ret_disc',
                    'ret_mavg_norm',
                    'ret_disc_norm'],
                legend_loc= 'lower left')

        dreturns = []
        if self.use_mavg:
            for rs in episode_rewards:
                dreturns += movavg_return(rewards=rs, factor=self.movavg_factor)
        else:
            for rs in episode_rewards:
                dreturns += discounted_return(rewards=rs, discount=self.discount)
        if self.do_zscore: dreturns = zscore_norm(dreturns)
        dreturns = np.array(dreturns, dtype=np.float32)

        out = self.actor.update_with_experience(
            observations=   observations,
            actions=        actions,
            dreturns=       dreturns,
            inspect=        inspect)

        value = out.pop('value').cpu().detach().numpy()
        advantage = out.pop('advantage').cpu().detach().numpy()

        if inspect:
            two_dim_multi(
                ys=     [
                    dreturns,
                    value,
                    advantage,
                ],
                names=  [
                    'dreturns',
                    'value',
                    'advantage',
                ],
                legend_loc= 'lower left')

        return out


"""

 2020 (c) piteren

    multiprocessing decorators >> run decorated functions in subprocess

    @proc
        - non-blocking
        - does not return any value

    @proc_wait
        - blocking

    @proc_return
        - blocking
        - returns function result (using que)

    @proc_que(que)
        - non-blocking
        - will put function result on given que

"""

from multiprocessing import Process, Queue
from functools import partial
import time

# non-blocking process, does not return anything
def proc(f):

    def new_f(*args, **kwargs):
        Process(target=partial(f, *args, **kwargs)).start()

    new_f.__name__ = f'@proc:{f.__name__}'
    return new_f

# blocking process, does not return anything
def proc_wait(f):

    def new_f(*args, **kwargs):
        p = Process(target=partial(f, *args, **kwargs))
        p.start()
        p.join()

    new_f.__name__ = f'@proc_wait:{f.__name__}'
    return new_f

# helper class (process with que - puts target result on que)
class MProc(Process):

    def __init__(
            self,
            que :Queue,
            f,
            args,
            kwargs):

        super().__init__(target=self.proc_m)
        self.que = que
        self.f = f
        self.ag = args
        self.kw = kwargs

    def proc_m(self):
        res = self.f(*self.ag, **self.kw)
        self.que.put(res)

# blocking process with return
def proc_return(f):
    def new_f(*args, **kwargs):
        que = Queue()
        p = MProc(que=que, f=f, args=args, kwargs=kwargs)
        p.start()
        ret = que.get()
        return ret
    new_f.__name__ = f'@proc_return:{f.__name__}'
    return new_f

# non-blocking process with return via que
def proc_que(que):
    def wrap(f):

        def new_f(*args, **kwargs):
            p = MProc(que=que, f=f, args=args, kwargs=kwargs)
            p.start()

        new_f.__name__ = f'@qproc:{f.__name__}'
        return new_f

    return wrap


# *********************************************************************** examples

def example_proc():

    @proc
    def task(to=5):
        for i in range(to):
            print(i, 'done')
            time.sleep(0.1)

    print(task.__name__)
    task(to=10)
    task(10)
    task(10)

    @proc
    def calc(n):
        val = 2 * n
        print(f'will return {val}')
        return val

    print(f'returned {calc(3)}')
    print(f'returned {calc(4)}')


def example_proc_wait():

    @proc_wait
    def task(to=5):
        for i in range(to):
            print(i, 'done')
            time.sleep(0.1)

    task(to=10)
    task(10)
    task(10)


def example_proc_return():
    @proc_return
    def calc(n):
        val = 2 * n
        print(f'will return {val}')
        return val

    print(f'returned {calc(3)}')
    print(f'returned {calc(4)}')


def example_proc_que():

    import random

    que = Queue()
    @proc_que(que)
    def task(name='def', to=5):
        sum = 0
        for i in range(to):
            print(f'calculating sum ({to})...')
            time.sleep(0.5)
            sum += i
        return name, sum

    print(task.__name__)

    n = 50
    for i in range(n): task(name=f'task_{i}', to=random.randrange(5,20))
    for _ in range(n): print(f'calculated result: {que.get()}')


if __name__ == '__main__':

    example_proc()
    #example_proc_wait()
    #example_proc_return()
    #example_proc_que()


"""
 2022 (c) piteren

 MOTorch implements NNWrap interface with PyTorch.

    nngraph for MOTorch:
        - Should be type of Module (defined below).
        - Should implement forward() and loss_acc() methods.

 MOTorch extends Module. By default, after init, MOTorch is set to train.mode=False.
 MOTorch manages its train.mode by itself.
"""

from abc import abstractmethod, ABC
import numpy as np
import shutil
from sklearn.metrics import f1_score
import torch
from typing import Optional, Tuple, Dict

from pypaq.lipytools.little_methods import stamp
from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.moving_average import MovAvg
from pypaq.lipytools.pylogger import get_hi_child
from pypaq.pms.parasave import ParaSave
from pypaq.comoneural.nnwrap import NNWrap, NNWrapException
from pypaq.mpython.devices import get_devices
from pypaq.torchness.types import TNS, DTNS
from pypaq.torchness.base_elements import mrg_ckpts
from pypaq.torchness.scaled_LR import ScaledLR
from pypaq.torchness.grad_clipping import GradClipperAVT


# torch.nn.Module to be implemented with forward & loss methods
# device/s are managed by MOTorch
class Module(ABC, torch.nn.Module):

    def __init__(self):
        torch.nn.Module.__init__(self)

    # returned dict should have at least 'logits' key with logits tensor
    @abstractmethod
    def forward(self, *args, **kwargs) -> DTNS:
        # return {'logits': self.logits(input)}
        raise NotImplementedError

    # baseline accuracy implementation for logits & lables
    def accuracy(
            self,
            logits: TNS,
            labels: TNS) -> float:
        logits = logits.detach().cpu().numpy()
        preds = np.argmax(logits, axis=-1)
        labels = labels.cpu().numpy()
        return float(np.average(np.equal(preds, labels)))

    # baseline F1 implementation for logits & lables
    def f1(self,
           logits: TNS,
           labels: TNS) -> float:
        logits = logits.detach().cpu().numpy()
        preds = np.argmax(logits, axis=-1)
        labels = labels.cpu().numpy()
        return f1_score(labels, preds, zero_division=0)

    # returned dict updates forward() Dict with loss & acc keys (accuracy or any other (increasing) performance float)
    @abstractmethod
    def loss_acc(self, *args, **kwargs) -> DTNS:
        # out = self.forward(input)
        # logits = out['logits']
        # out['loss'] = torch.nn.functional.cross_entropy(logits, labels, reduction='mean')
        # out['acc'] = self.accuracy(logits, labels)
        # out['f1'] = self.f1(logits, labels)
        raise NotImplementedError



class MOTorchException(NNWrapException):
    pass


# extends Module (torch.nn.Module) with ParaSave and many others
class MOTorch(NNWrap, torch.nn.Module):

    SPEC_KEYS = {
        'loss',         # loss
        'acc',          # accuracy
        'f1'}           # F1

    INIT_DEFAULTS = {
        'seed':             123,                # seed for torch and numpy
        'devices':          -1,                 # :DevicesParam (check pypaq.mpython.devices)
        'dtype':            torch.float32,
            # training
        'batch_size':       64,                 # training batch size
        'n_batches':        1000,               # default length of training
        'opt_class':        torch.optim.Adam,   # default optimizer
        'train_batch_IX':   0,                  # default (starting) batch index (counter)
            # LR management (check pypaq.torchness.base_elements.ScaledLR)
        'baseLR':           3e-4,
        'warm_up':          None,
        'ann_base':         None,
        'ann_step':         1.0,
        'n_wup_off':        2.0,
            # gradients clipping parameters (check pypaq.torchness.base_elements.GradClipperAVT)
        'clip_value':       None,
        'avt_SVal':         0.1,
        'avt_window':       100,
        'avt_max_upd':      1.5,
        'do_clip':          False,
            # other
        'hpmser_mode':      False,              # it will set model to be read_only and quiet when running with hpmser
        'read_only':        False,              # sets model to be read only - wont save anything (wont even create self.nnwrap_dir)
        'do_TB':            True}               # runs TensorBard, saves in self.nnwrap_dir

    SAVE_FN_PFX = 'motorch_dna' # filename (DNA) prefix

    def __init__(
            self,
            nngraph: Optional[type(Module)]=    None,
            name: Optional[str]=                None,
            name_timestamp=                     False,
            save_topdir: Optional[str]=         None,
            save_fn_pfx: Optional[str]=         None,
            logger=                             None,
            loglevel=                           20,
            **kwargs):

        torch.nn.Module.__init__(self)

        self._torch_dev = None # will be set by _manage_devices()
        self._opt = None
        self._scheduler = None
        self._grad_clipper = None

        self._nngraph_module: nngraph = None

        NNWrap.__init__(
            self,
            name=           name,
            name_timestamp= name_timestamp,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx,
            logger=         logger,
            loglevel=       loglevel,
            nngraph=        nngraph,
            **kwargs)

    # ******************************************************************************************* NNWrap init submethods

    def _generate_name(
            self,
            given_name: Optional[str],
            timestamp: bool) -> str:
        name = f'{self.nngraph.__name__}_MOTorch' if not given_name else given_name
        if timestamp: name += f'_{stamp()}'
        return name

    # sets CPU / GPU devices for MOTorch
    def _manage_devices(self):

        self._nwwlog.debug(f'> MOTorch resolves devices, given: {self["devices"]}, torch.cuda.is_available(): {torch.cuda.is_available()}')

        dev = get_devices(
            devices=    self['devices'],
            namespace=  'torch',
            logger=     get_hi_child(self._nwwlog, 'get_devices'))
        self._nwwlog.info(f'> MOTorch given devices: {self["devices"]}, will use {dev}')
        # TODO: by now supported is only the first given device
        self._torch_dev = torch.device(dev[0])

    # sets seed in all possible areas
    def _set_seed(self) -> None:
        # seed (https://pytorch.org/docs/stable/notes/randomness.html)
        torch.manual_seed(self['seed'])
        torch.cuda.manual_seed(self['seed'])
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    # builds MOTorch graph (Module)
    def _build_graph(self) -> None:

        self._nwwlog.info('MOTorch builds graph')
        self._nngraph_module = self.nngraph(**self._dna_nngraph)
        self.to(self._torch_dev)
        self.to(self['dtype'])
        self._nwwlog.debug(f'{self.name} (MOTorch) Module initialized!')

        try:
            self.load_ckpt()
            self._nwwlog.info(f'> MOTorch checkpoint loaded from {self.__get_ckpt_path()}')
        except Exception as e:
            self._nwwlog.info(f'> MOTorch checkpoint NOT loaded ({e})..')

        # optimizer params may be given with 'opt_' prefix
        opt_params = {k[4:]: self[k] for k in self.get_managed_params() if k.startswith('opt_')}
        opt_params.pop('class')
        self._opt = self['opt_class'](
            params= self.parameters(),
            lr=     self['baseLR'],
            **opt_params)
        #print(len(self._opt.param_groups))
        #print(self._opt.param_groups[0].keys())

        # from now LR is managed by scheduler
        self._scheduler = ScaledLR(
            optimizer=      self._opt,
            starting_step=  self['train_batch_IX'],
            warm_up=        self['warm_up'],
            ann_base=       self['ann_base'],
            ann_step=       self['ann_step'],
            n_wup_off=      self['n_wup_off'],
            logger=         get_hi_child(self._nwwlog, 'ScaledLR'))

        self._grad_clipper = GradClipperAVT(
            module=         self,
            clip_value=     self['clip_value'],
            avt_SVal=       self['avt_SVal'],
            avt_window=     self['avt_window'],
            avt_max_upd=    self['avt_max_upd'],
            do_clip=        self['do_clip'],
            logger=         get_hi_child(self._nwwlog, 'GradClipperAVT'))

        self.train(False)
        self._nwwlog.debug(f'> set MOTorch train.mode to False..')

    def __call__(self, *args, **kwargs) -> dict:
        return torch.nn.Module.__call__(self, *args, **kwargs)

    # *********************************************************************************************** load / save / copy

    # returns path of checkpoint pickle file
    @staticmethod
    def __get_ckpt_path_static(model_dir:str, model_name:str) -> str:
        return f'{model_dir}/{model_name}.pt'

    # returns path of checkpoint pickle file
    def __get_ckpt_path(self) -> str:
        return self.__get_ckpt_path_static(self.nnwrap_dir, self.name)

    def load_ckpt(self) -> None:
        # TODO: load all that has been saved
        """
        print(f'@@@@@ {self._torch_dev}')
        INFO: for OSX (mac) self._torch_dev == 'cpu:0' and it does not load, since:
        MOTorch checkpoint NOT loaded: don't know how to restore data location of torch.storage.UntypedStorage (tagged with cpu:0)
        """
        checkpoint = torch.load(
            self.__get_ckpt_path(),
            #map_location=   'cpu') # works for for OSX
            map_location=   self._torch_dev) # INFO: to immediately place all tensors to current device (not previously saved one)
        self.load_state_dict(checkpoint['model_state_dict'])

    # saves model checkpoint
    def save_ckpt(self) -> None:
        # TODO: decide what to save
        torch.save({
            #'epoch': 5,
            'model_state_dict': self.state_dict(),
            # 'optimizer_state_dict': optimizer.state_dict(),
            #'loss': 0.4
        }, self.__get_ckpt_path())

    @classmethod
    def copy_checkpoint(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None):
        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_topdir_trg: save_topdir_trg = save_topdir_src
        shutil.copyfile(
            src=    MOTorch.__get_ckpt_path_static(f'{save_topdir_src}/{name_src}', name_src),
            dst=    MOTorch.__get_ckpt_path_static(f'{save_topdir_trg}/{name_trg}', name_trg))

    # *************************************************************************************************************** GX

    @classmethod
    def gx_ckpt(
            cls,
            name_A: str,                            # name parent A
            name_B: str,                            # name parent B
            name_child: str,                        # name child
            save_topdir_A: Optional[str]=       None,
            save_topdir_B: Optional[str]=       None,
            save_topdir_child: Optional[str]=   None,
            ratio: float=                       0.5,
            noise: float=                       0.03):

        if not save_topdir_A: save_topdir_A = cls.SAVE_TOPDIR
        if not save_topdir_B: save_topdir_B = save_topdir_A
        if not save_topdir_child: save_topdir_child = save_topdir_A

        model_dir_child = f'{save_topdir_child}/{name_child}'
        prep_folder(model_dir_child)

        mrg_ckpts(
            ckptA=          MOTorch.__get_ckpt_path_static(f'{save_topdir_A}/{name_A}', name_A),
            ckptB=          MOTorch.__get_ckpt_path_static(f'{save_topdir_B}/{name_B}', name_B),
            ckptM=          MOTorch.__get_ckpt_path_static(model_dir_child, name_child),
            ratio=          ratio,
            noise=          noise)

    # ***************************************************************************************************** train / test

    # converts all values given with args & kwargs to tensors and moves to self._torch_dev (device)
    def __torch_dev(
            self,
            *args,
            to_torch=   True,  # converts given data to torch.Tensors
            to_devices= True,  # moves tensors to devices
            **kwargs):
        if to_torch:
            args = [torch.tensor(a) if type(a) is not torch.Tensor else a for a in args]
            kwargs = {k: torch.tensor(kwargs[k]) if type(kwargs[k]) is not torch.Tensor else kwargs[k] for k in kwargs}
        if to_devices:
            args = [a.to(self._torch_dev) for a in args]
            kwargs = {k: kwargs[k].to(self._torch_dev) for k in kwargs}
        return args, kwargs

    # runs forward on nn.Module (with current nn.Module.training.mode - by default not training)
    # INFO: since MOTorch is a torch.nn.Module, call forward() call should be avoided, instead use just MOTorch.__call__() /self()
    def forward(
            self,
            *args,
            to_torch=                       True,   # converts given data to torch.Tensors
            to_devices=                     True,   # moves tensors to devices
            set_training: Optional[bool]=   None,   # for not None forces given training mode for torch.nn.Module
            **kwargs) -> DTNS:
        if set_training is not None: self.train(set_training)
        args, kwargs = self.__torch_dev(*args, to_torch=to_torch, to_devices=to_devices, **kwargs)
        out = self._nngraph_module.forward(*args, **kwargs)
        if set_training: self.train(False) # eventually roll back to default
        return out

    # runs loss calculation on nn.Module (with current nn.Module.training.mode - by default not training)
    def loss_acc(
            self,
            *args,
            to_torch=                       True,   # converts given data to torch.Tensors
            to_devices=                     True,   # moves tensors to devices
            set_training: Optional[bool]=   None,   # for not None forces given training mode for torch.nn.Module
            **kwargs) -> DTNS:
        if set_training is not None: self.train(set_training)
        args, kwargs = self.__torch_dev(*args, to_torch=to_torch, to_devices=to_devices, **kwargs)
        out = self._nngraph_module.loss_acc(*args, **kwargs)
        if set_training: self.train(False) # eventually roll back to default
        return out

    # runs loss calculation + update of nn.Module (by default with training.mode = True)
    def backward(
            self,
            *args,
            to_torch=           True,   # converts given data to torch.Tensors
            to_devices=         True,   # moves tensors to devices
            set_training: bool= True,
            **kwargs) -> DTNS:

        out = self.loss_acc(
            *args,
            to_torch=       to_torch,
            to_devices=     to_devices,
            set_training=   set_training,
            **kwargs)

        out['loss'].backward()          # update gradients
        gnD = self._grad_clipper.clip() # clip gradients, adds: 'gg_norm' & 'gg_avt_norm' to out
        self._opt.step()                # apply optimizer
        self._opt.zero_grad()           # clear gradients
        self._scheduler.step()          # apply LR scheduler

        out['currentLR'] = self._scheduler.get_last_lr()[0] # INFO: we take currentLR of first group
        out.update(gnD)

        return out

    # adds conversion to torch.Tensors and moves to device
    def load_data(
            self,
            data_TR: Dict,
            data_VL: Optional[Dict]=    None,
            data_TS: Optional[Dict]=    None):

        data = {
            'data_TR': data_TR,
            'data_VL': data_VL,
            'data_TS': data_TS}
        data_td = {}
        for k in data:
            if data[k]:
                data_td[k] = {}
                for l in data[k]:
                    d = data[k][l]
                    if type(d) is not torch.Tensor: d = torch.tensor(d)
                    d = d.to(self._torch_dev)
                    data_td[k][l] = d

        super(MOTorch, self).load_data(**data_td)

    def run_train(
            self,
            n_batches: Optional[int]=   None,
            test_freq=                  100,
            mov_avg_factor=             0.1,
            save_max=                   True,
            use_F1=                     True,
            **kwargs) -> Optional[float]:

        if not self._batcher: raise MOTorchException('MOTorch has not been given data for training, use load_data()')

        self._nwwlog.info(f'{self.name} - training starts [acc / F1 / loss]')
        self._nwwlog.info(f'data sizes (TR,VL,TS) samples: {self._batcher.get_data_size()}')

        if n_batches is None: n_batches = self['n_batches']  # take default
        self._nwwlog.info(f'batch size:             {self["batch_size"]}')
        self._nwwlog.info(f'train for num_batches:  {n_batches}')

        self.train(True)

        batch_IX = 0                            # this loop (local) batch counter
        tr_accL = []
        tr_f1L = []
        tr_lssL = []

        score_name = 'F1' if use_F1 else 'acc'
        ts_score_max = 0                        # test score (acc or F1) max
        ts_score_all_results = []               # test score all results
        ts_score_mav = MovAvg(mov_avg_factor)   # test score (acc or F1) moving average

        ts_bIX = [bIX for bIX in range(n_batches+1) if not bIX % test_freq] # batch indexes when test will be performed
        assert ts_bIX, 'ERR: model SHOULD BE tested while training!'
        ten_factor = int(0.1*len(ts_bIX)) # number of tests for last 10% of training
        if ten_factor < 1: ten_factor = 1 # we need at least one result
        if self['hpmser_mode']: ts_bIX = ts_bIX[-ten_factor:]

        while batch_IX < n_batches:

            out = self.backward(
                to_torch=   False,
                to_devices= False,
                **self._batcher.get_batch())

            acc = out['acc'] if 'acc' in out else None
            f1 = out['f1'] if 'f1' in out else None

            batch_IX += 1
            self['train_batch_IX'] += 1

            if self['do_TB']:
                self.log_TB(value=out['loss'],          tag='tr/loss',      step=self['train_batch_IX'])
                self.log_TB(value=out['gg_norm'],       tag='tr/gn',        step=self['train_batch_IX'])
                self.log_TB(value=out['gg_avt_norm'],   tag='tr/gn_avt',    step=self['train_batch_IX'])
                self.log_TB(value=out['currentLR'],     tag='tr/cLR',       step=self['train_batch_IX'])
                if acc is not None:
                    self.log_TB(value=acc,              tag='tr/acc',       step=self['train_batch_IX'])
                if f1 is not None:
                    self.log_TB(value=f1,               tag='tr/F1',       step=self['train_batch_IX'])

            if acc is not None: tr_accL.append(acc)
            if f1 is not None: tr_f1L.append(f1)
            tr_lssL.append(out['loss'])

            if batch_IX in ts_bIX:

                self.train(False)
                ts_acc, ts_f1, ts_loss = self.run_test()
                self.train(True)

                ts_score = ts_f1 if use_F1 else ts_acc
                if ts_score is not None:
                    ts_score_all_results.append(ts_score)
                if self['do_TB']:
                    self.log_TB(value=ts_loss,                          tag='ts/loss',              step=self['train_batch_IX'])
                    if ts_acc is not None:
                        self.log_TB(value=ts_acc,                       tag='ts/acc',               step=self['train_batch_IX'])
                    if ts_f1 is not None:
                        self.log_TB(value=ts_f1,                        tag='ts/F1',                step=self['train_batch_IX'])
                    if ts_score is not None:
                        self.log_TB(value=ts_score_mav.upd(ts_score),   tag=f'ts/{score_name}_mav', step=self['train_batch_IX'])

                tr_acc_nfo = f'{100*sum(tr_accL)/test_freq:.1f}' if acc is not None else '--'
                tr_f1_nfo =  f'{100*sum(tr_f1L)/test_freq:.1f}' if f1 is not None else '--'
                ts_acc_nfo = f'{100*ts_acc:.1f}' if ts_acc is not None else '--'
                ts_f1_nfo = f'{100*ts_f1:.1f}' if ts_f1 is not None else '--'
                self._nwwlog.info(f'# {self["train_batch_IX"]:5d} TR: {tr_acc_nfo} / {tr_f1_nfo} / {sum(tr_lssL)/test_freq:.3f} -- TS: {ts_acc_nfo} / {ts_f1_nfo} / {ts_loss:.3f}')
                tr_accL = []
                tr_f1L = []
                tr_lssL = []

                if ts_score is not None and ts_score > ts_score_max:
                    ts_score_max = ts_score
                    if not self['read_only'] and save_max: self.save_ckpt() # model is saved for max ts_score

        # weighted (linear ascending weight) test score for last 10% test results
        ts_score_wval = None
        if ts_score_all_results:
            ts_score_wval = 0.0
            weight = 1
            sum_weight = 0
            for tr in ts_score_all_results[-ten_factor:]:
                ts_score_wval += tr*weight
                sum_weight += weight
                weight += 1
            ts_score_wval /= sum_weight

            if self['do_TB']: self.log_TB(value=ts_score_wval, tag=f'ts/ts_{score_name}_wval', step=self['train_batch_IX'])

        self._nwwlog.info(f'### model {self.name} finished training')
        if ts_score_wval is not None:
            self._nwwlog.info(f' > test_{score_name}_max:  {ts_score_max:.4f}')
            self._nwwlog.info(f' > test_{score_name}_wval: {ts_score_wval:.4f}')

        self.train(False)

        return ts_score_wval

    def run_test(
            self,
            set_training: Optional[bool]=   None,   # for not None sets training mode for torch.nn.Module, allows to calculate loss with training/evaluating module mode
    ) -> Tuple[Optional[float], Optional[float], float]:

        if not self._batcher: raise MOTorchException('MOTorch has not been given data for testing, use load_data() or give it while testing!')

        if set_training is not None: self.train(set_training)

        batches = self._batcher.get_TS_batches()
        lossL = []
        accL = []
        f1L = []
        for batch in batches:
            out = self.loss_acc(**batch)
            lossL.append(out['loss'])
            if 'acc' in out:
                accL.append(out['acc'])
            if 'f1' in out:
                f1L.append(out['f1'])

        if set_training is not None: self.train(False)  # eventually roll back to default

        acc_avg = sum(accL)/len(accL) if accL else None
        f1_avg = sum(f1L)/len(f1L) if f1L else None
        return acc_avg, f1_avg, sum(lossL)/len(lossL)

    # updates scheduler baseLR of 0 group
    def update_baseLR(self, lr: float):
        self['baseLR'] = lr # in case model will be saved >> loaded
        self._scheduler.update_base_lr0(lr)

    def __str__(self):
        s = f'MOTorch: {ParaSave.__str__(self)}\n'
        s += str(self._nngraph_module)
        return s


from pypaq.neuralmess.base_elements import tf
from pypaq.neuralmess.layers import lay_dense

def acs_graph(
        name=               'acs',
        observation_width=  4,
        num_actions=        2,
        hidden_layers=      (24,24),
        lay_norm=           False,
        seed=               123,
        verb=               0):

    with tf.variable_scope(name):

        observation_PH = tf.placeholder(
            shape=  (None, observation_width),
            dtype=  tf.float32,
            name=   'observation')

        action_PH = tf.placeholder( # action taken (may differ from Actor policy action because of Trainer policy like exploration)
            shape=  None,
            dtype=  tf.int32,
            name=   'action')

        qv_label_PH = tf.placeholder( # label of Q(s,a), computed from: reward + gamma*V_next_action
            shape=  None,
            dtype=  tf.float32,
            name=   'next_action_qvs')

        layer = observation_PH
        if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)
        for i in range(len(hidden_layers)):
            layer = lay_dense(
                input=      layer,
                name=       f'hidden_layer_{i}',
                units=      hidden_layers[i],
                activation= tf.nn.relu,
                seed=       seed)
            if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)

        # critic QVs (for all actions of current observation)
        qvs = lay_dense(
            input=      layer,
            name=       'qvs',
            units=      num_actions,
            activation= None,
            seed=       seed)

        # action logits of Actor policy
        action_logits = lay_dense(
            input=      layer,
            name=       'action_logits',
            units=      num_actions,
            activation= None,
            seed=       seed)
        action_prob = tf.nn.softmax(action_logits)

        value = tf.reduce_sum(qvs * action_prob)  # value of observation (next_observation)

        max_probs = tf.reduce_max(action_prob, axis=-1) # max action_probs
        min_probs = tf.reduce_min(action_prob, axis=-1) # min action_probs
        amax_prob = tf.reduce_mean(max_probs) # average of batch max action_prob
        amin_prob = tf.reduce_mean(min_probs) # average of batch min action_prob

        action_OH = tf.one_hot(action_PH, num_actions)
        qv = tf.reduce_sum(qvs*action_OH, axis=1) # Q(s,a)
        log_policy = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=action_logits, labels=action_PH)
        loss_actor = tf.reduce_mean(qv * log_policy)

        #next_V = tf.reduce_sum(next_action_qvs_PH * next_action_probs_PH, axis=1) # V(next_s)
        #labels = reward_PH + gamma * next_V
        # TODO: try with hubner loss
        loss_critic = tf.reduce_mean(tf.losses.mean_squared_error(predictions=qv, labels=qv_label_PH))

        loss = loss_actor + loss_critic

        if verb>0:
            print('AC_shared graph debug:')
            print(f' > qvs: {qvs}')
            print(f' > action_logits: {action_logits}')
            print(f' > value: {value}')

    return {
        'observation_PH':   observation_PH,
        'action_PH':        action_PH,
        'qv_label_PH':      qv_label_PH,

        'qvs':              qvs,
        'action_prob':      action_prob,
        'value':            value,

        'amax_prob':        amax_prob,
        'amin_prob':        amin_prob,

        'loss_actor':       loss_actor,
        'loss_critic':      loss_critic,
        'loss':             loss}


"""

 2022 (c) piteren

    DQNPTActor - DQNActor implemented with MOTorch (PyTorch)

"""

import numpy as np
from typing import List, Optional

from pypaq.R4C.qlearning.dqn.dqn_actor import DQN_Actor
from pypaq.R4C.qlearning.dqn.pt_based.dqn_PT_module import DQNModel
from pypaq.torchness.motorch import MOTorch, Module


# DQN (PyTorch NN) based QLearningActor
class DQN_PTActor(DQN_Actor):

    def __init__(
            self,
            name: str=                          'DQN_PTActor',
            nngraph: Optional[type(Module)]=    DQNModel,
            **kwargs):
        DQN_Actor.__init__(
            self,
            name=       name,
            nnwrap=     MOTorch,
            nngraph=    nngraph,
            **kwargs)

    def _get_QVs(self, observation: object) -> np.ndarray:
        obs_vec = self._get_observation_vec(observation)
        return self.nnw(obs_vec)['logits'].detach().cpu().numpy()

    # optimized with single call with a batch of observations
    def get_QVs_batch(self, observations: List[object]) -> np.ndarray:
        obs_vecs = np.array([self._get_observation_vec(o) for o in observations])
        return self.nnw(obs_vecs)['logits'].detach().cpu().numpy()

    # optimized with single call to session with a batch of data
    def update_with_experience(
            self,
            observations: List[object],
            actions: List[int],
            new_qvs: List[float],
            inspect=    False) -> dict:

        obs_vecs = self._get_observation_vec_batch(observations)
        full_qvs = np.zeros_like(obs_vecs)
        mask = np.zeros_like(obs_vecs)
        for v,pos in zip(new_qvs, enumerate(actions)):
            full_qvs[pos] = v
            mask[pos] = 1

        self._rlog.log(5, f'>> obs_vecs (shape {obs_vecs.shape})\n{obs_vecs}')
        self._rlog.log(5, f'>> actions (len {len(actions)}): {actions}')
        self._rlog.log(5, f'>> new_qvs (len {len(new_qvs)}): {new_qvs}')
        self._rlog.log(5, f'>> full_qvs\n{full_qvs}')
        self._rlog.log(5, f'>> mask\n{mask}')

        out = self.nnw.backward(obs_vecs, full_qvs, mask)
        out.pop('logits')
        out.pop('acc') # accuracy for DQN does not make sense
        return out



from typing import List

# returns nice string of floats list
def _str_weights(
        all_w :List[float],
        cut_above=      5,
        float_prec=     4) -> str:
    ws = '['
    if cut_above < 5: cut_above = 5 # cannot be less than 5
    if len(all_w) > cut_above:
        for w in all_w[:3]: ws += f'{w:.{float_prec}f} '
        ws += '.. '
        for w in all_w[-2:]: ws += f'{w:.{float_prec}f} '
    else:
        for w in all_w: ws += f'{w:.{float_prec}f} '
    return f'{ws[:-1]}]'


import numpy as np
from typing import List



# ReinforcementLearning Exception
class RLException(Exception): pass


# extracts array of data from a batch
def extract_from_batch(
        batch :List[dict],
        key: str) -> List:
    return list(map(lambda x: x[key], batch))

# normalizes x with zscore (0 mean 1 std), this is helpful for training, as rewards can vary considerably between episodes,
def zscore_norm(x):
    if len(x) < 2: return x
    return (x - np.mean(x)) / (np.std(x) + 0.00000001)

# prepares list of discounted accumulated return from [reward]
def discounted_return(
        rewards: List[float],
        discount: float) -> List[float]:
    dar = np.zeros_like(rewards)
    s = 0.0
    for i in reversed(range(len(rewards))):
        s = s * discount + rewards[i]
        dar[i] = s
    return list(dar)

# prepares list of moving_average return from [reward]
def movavg_return(
        rewards: List[float],
        factor: float           # (0.0-0.1> factor of current reward taken for update
) -> List[float]:
    mvr = np.zeros_like(rewards)
    s = rewards[-1]
    mvr[-1] = s
    for i in reversed(range(len(rewards[:-1]))):
        s = (1-factor) * s + factor * rewards[i]
        mvr[i] = s
    return list(mvr)


"""

 2020 (c) piteren

    hpmser - hyperparameters searching function
        > searches hyperparameters space to MAXIMIZE the SCORE of func

        MAXIMIZE the SCORE == find points cluster with:
         - high num of points in a small distance (dst)
         - high max smooth_score
         - high lowest score

         policy of sampling the space is crucial (determines the speed, top result and convergence)
         - fully random sampling is slow, wastes a lot of time and computing power
         - too aggressive sampling may undersample the space and miss the real MAX

    to setup search:
    1. having some FUNCTION:
        - some parameters need to be optimized
        - some parameters may be fixed / constant
        - if function accepts 'device' or 'devices' it should be type of DevicesParam (check pypaq.mpython.devices),
          it will be used by hpmser to put proper device for each function call
        - returns a dict with 'score' or just a value (score)
        There are two parameters of FUNCTION that may be used by hpmser:
            - 'devices' (type of DevicesParam - check pypaq.mpython.devices) -> hpmser will send devices to FUNCTION
            - 'hpmser_mode' -> will be set to True by hpmser
    2. define PSDD - dictionary with parameters to be optimized and the space to search in (check pypaq.pms.pasap.PaSpa)
    3. import hpmser function into your script, and run it with:
        - func << FUNCTION
        - func_psdd << PSDD
        - func_const << dictionary of parameters that have to be given but should be fixed / constant durring optimization
        - .. other params configure hpmser algorithm itself
"""

#TODO:
# - look for solution for exponential complexity of space management functions
# - build hpmser multiserver
# - limited number of permutations for given PSDD (not-continuous space search case)
# - add option to search for minimum rather than maximum


import os
import sys, select
import time
from typing import Callable, Optional, List, Any

from pypaq.hpmser.search_results import SRL
from pypaq.hpmser.helpers import _str_weights
from pypaq.lipytools.little_methods import stamp, get_params
from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.logger import set_logger
from pypaq.lipytools.stats import msmx
from pypaq.mpython.devices import DevicesParam, get_devices
from pypaq.mpython.omp import OMPRunner, RunningWorker
from pypaq.torchness.base_elements import TBwr
from pypaq.pms.config_manager import ConfigManager
from pypaq.pms.paspa import PaSpa
from pypaq.pms.base_types import PSDD, POINT, point_str


NP_SMOOTH = [3,5,9] # numbers of points for smoothing

# initial hpmser configuration
SAMPLING_CONFIG_INITIAL = {
    'np_smooth':    3,      # number of points used for smoothing
    'prob_opt':     0.3,    # probability of sampling from estimated space (optimized)
    'n_opt':        30,     # number of samples taken from estimated space
    'prob_top':     0.3,    # probability of sampling from the area of top points
    'n_top':        20}     # number of samples taken from area of top points

# hpmser configuration updated values - hpmser will be automatically updated after 'config_upd' loops
SAMPLING_CONFIG_UPD = {
    'prob_opt':     0.4,
    'prob_top':     0.6}


# hyper-parameters searching function (based on OMP engine)
#@timing <- temporary disabled
def hpmser(
        func: Callable,                                 # function which parameters need to be optimized
        func_psdd: PSDD,                                # func PSDD, from here points {param: arg} will be sampled
        func_const: Optional[POINT]=    None,           # func constant kwargs, will be updated with sample (point) taken from PaSpa
        devices: DevicesParam=          None,           # devices to use for search
        name: Optional[str]=            None,           # hpmser run name, for None stamp will be used
        add_stamp=                      True,           # adds short stamp to name, when name given
        use_GX=                         True,           # uses genetic xrossing while sampling top points
        distance_L2=                    True,           # use L2(True) or L1(False) for distance calculation
        stochastic_est: Optional[int]=  3,              # number of samples used for stochastic estimation, for 0 or None does not estimate
        config_upd: Optional[int]=      1000,           # update config after n loops
        n_loops: Optional[int]=         None,           # limit for number of search loops
        hpmser_FD: str=                 '_hpmser_runs', # save folder
        do_log=                         True,           # sets logger
        do_TB=                          True,           # plots with TB
        pref_axes: Optional[List[str]]= None,           # preferred axes for plot, put here a list of up to 3 params names ['param1',..]
        top_show_freq=                  20,             # how often top results summary will be printed
        restart_exceptions_tasks=       False,          # does not restart task that raised exception
        raise_exceptions=               True,           # forces subprocesses to raise + print exceptions, independent from verbosity (raising subprocess exception does not break hpmser process)
        verb=                           1) -> SRL:

    # hpmser RunningWorker (process run by OMP in hpmser)
    class HRW(RunningWorker):

        def __init__(
                self,
                func: Callable,
                func_const: Optional[POINT],
                devices: DevicesParam = None):

            self.func = func
            self.func_const = func_const if func_const else {}
            self.devices = devices

            # manage 'devices' & 'hpmser_mode' param in func >> set it in func if needed
            func_args = get_params(self.func)
            func_args = list(func_args['with_defaults'].keys()) + func_args['without_defaults']
            assert 'device' not in func_args, f'ERR: you should use "devices" instead of "device" in your "{self.func.__name__}()" function'
            if 'devices' in func_args: self.func_const['devices'] = self.devices
            if 'hpmser_mode' in func_args: self.func_const['hpmser_mode'] = True

        # processes given spoint, passes **kwargs
        def process(
                self,
                spoint: POINT,
                **kwargs) -> Any:

            spoint_with_defaults = {}
            spoint_with_defaults.update(self.func_const)
            spoint_with_defaults.update(spoint)

            res = self.func(**spoint_with_defaults)
            if type(res) is dict: score = res['score']
            else:                 score = res

            msg = {'spoint':spoint, 'score':score}
            msg.update(kwargs)
            return msg

    # some defaults
    if not name: name = stamp()
    elif add_stamp: name = f'{stamp(letters=0)}_{name}'

    prep_folder(hpmser_FD)  # create folder if needed

    # check for continuation
    srl = None
    results_FDL = sorted(os.listdir(hpmser_FD))
    if len(results_FDL):
        if verb>0: print(f'\nThere are {len(results_FDL)} searches in \'{hpmser_FD}\', do you want to continue with the last one ({results_FDL[-1]})? .. waiting 10 sec (y/n, n-default)')
        i, o, e = select.select([sys.stdin], [], [], 10)
        if i and sys.stdin.readline().strip() == 'y':
            name = results_FDL[-1]  # take last
            srl = SRL(name=name, verb=verb)
            srl.load(f'{hpmser_FD}/{name}')
            assert PaSpa(psdd=func_psdd, distance_L2=distance_L2) == srl.paspa, 'ERR: parameters space differs - cannot continue!'

    if do_log: set_logger(log_folder=f'{hpmser_FD}/{name}', custom_name=name, add_stamp=False, verb=verb)  # set logger

    prep_folder(f'{hpmser_FD}/{name}') # needs to be created for ConfigManager
    config_manager = ConfigManager(
        file=           f'{hpmser_FD}/{name}/hpmser.conf',
        config=         SAMPLING_CONFIG_INITIAL,
        try_to_load=    True)
    sampling_config = config_manager.get_config() # update in case of read from existing file

    tbwr = TBwr(logdir=f'{hpmser_FD}/{name}') if do_TB else None

    if verb>0:
        print(f'\n*** hpmser {name} *** started for: {func.__name__}, sampling config: {sampling_config}')
        if srl: print(f' search will continue with {len(srl)} results...')

    if not srl: srl = SRL(
        paspa=  PaSpa(
            psdd=           func_psdd,
            distance_L2=    distance_L2,
            verb=           verb-1),
        name=   name,
        verb=   verb)
    srl.plot_axes = pref_axes

    if verb>0: print(f'\n{srl.paspa}')

    # prepare special points: corners and stochastic
    cpa, cpb = srl.paspa.sample_corners()
    special_points = {0:cpa, 1:cpb}
    if stochastic_est is None: stochastic_est = 0
    stochastic_points = {}
    stochastic_results = []

    avg_dst = srl.get_avg_dst()
    sample_num = len(srl) # number of next sample that will be taken and sent for processing

    curr_max_sr_id = None if not len(srl) else srl.get_top_SR().id  # current max SeRes id
    prev_max_sr_id = None                                           # previous max SeRes id

    scores_all = []

    devices = get_devices(devices=devices, namespace=None) # manage devices

    num_free_rw = len(devices)

    omp = OMPRunner(
        rw_class=               HRW,
        rw_init_kwargs=         {'func': func, 'func_const':func_const},
        rw_lifetime=            1,
        devices=                devices,
        name=                   'OMPR_NB_Hpmser',
        ordered_results=        False,
        restart_ex_tasks=       restart_exceptions_tasks,
        log_exceptions=         verb > 0 or raise_exceptions,
        raise_RWW_exception=    verb>1 or raise_exceptions,
        verb=                   verb-1)

    top_time = time.time()
    top_speed_save = []
    print(f'\nhpmser starts search loop..\n -- id smooth [local diff_VS_est] topID:dist_to avg_distance/max_of_min_distances time')
    try:
        while True:

            new_sampling_config = config_manager.load()
            if new_sampling_config: sampling_config.update(new_sampling_config)
            if 'np_smooth' in new_sampling_config: srl.set_np_smooth(sampling_config['np_smooth'])

            # use all available devices
            while num_free_rw:
                if verb>1: print(f' >> got {num_free_rw} free RW at {sample_num} sample_num start')

                if config_upd == sample_num:
                    if verb>0: print(f' > updating sampling config..')
                    new_sampling_config = config_manager.update(**SAMPLING_CONFIG_UPD)
                    sampling_config.update(new_sampling_config)

                # fill stochastic points after 50 samples
                if sample_num == 50:
                    _sp = srl.get_top_SR().point                        # try with top
                    if _sp is None: _sp = srl.paspa.sample_point_GX()   # else take random
                    for ix in range(stochastic_est):
                        stochastic_points[sample_num+ix] = _sp

                spoint = None
                est_score = 0

                # use special_points
                if sample_num in special_points:    spoint = special_points[sample_num]
                if sample_num in stochastic_points: spoint = stochastic_points[sample_num]
                # or take point with policy
                if not spoint:
                    params = {
                        'prob_opt': sampling_config['prob_opt'],
                        'n_opt':    sampling_config['n_opt'],
                        'prob_top': sampling_config['prob_top'],
                        'n_top':    sampling_config['n_top'],
                        'avg_dst':  avg_dst}
                    spoint, est_score = srl.get_opt_sample_GX(**params) if use_GX else srl.get_opt_sample(**params)

                task = {
                    'spoint':           spoint,
                    'sample_num':       sample_num,
                    'est_score':        est_score,
                    's_time':           time.time()}

                omp.process(task)
                num_free_rw -= 1
                sample_num += 1

            msg = omp.get_result(block=True) # get one result
            num_free_rw += 1
            if type(msg) is dict: # we may receive str here (like: 'TASK #4 RAISED EXCEPTION') from omp that not restarts exceptions

                msg_sample_num =    msg['sample_num']
                msg_score =         msg['score']
                msg_spoint =        msg['spoint']
                msg_est_score =     msg['est_score']
                msg_s_time =        msg['s_time']

                # manage stochastic estimation without adding to the results
                if msg_sample_num in stochastic_points:
                    stochastic_results.append(msg_score)
                    if len(stochastic_results) == stochastic_est and verb:
                        print(f'\n*** stochastic estimation with {stochastic_est} points:')
                        print(f'  > results: {_str_weights(stochastic_results, float_prec=8)}')
                        print(f'  > std_dev: {msmx(stochastic_results)["std"]:.8f}\n')

                else:
                    sr = srl.add_result(
                        point=  msg_spoint,
                        score=  msg_score)
                    if verb>1: print(f' >> got result #{msg_sample_num}')

                    pf = f'.{srl.prec}f' # update precision of print

                    avg_dst = srl.get_avg_dst()
                    mom_dst = srl.get_mom_dst()
                    srl.save(folder=f'{hpmser_FD}/{name}')

                    top_SR = srl.get_top_SR()

                    # gots new MAX
                    gots_new_max = False
                    if top_SR.id != curr_max_sr_id:
                        prev_max_sr_id = curr_max_sr_id
                        curr_max_sr_id = top_SR.id
                        gots_new_max = True

                    # current sr report
                    if verb>0:

                        dif = sr.smooth_score - msg_est_score
                        difs = f'{"+" if dif>0 else "-"}{abs(dif):{pf}}'

                        dist_to_max = srl.paspa.distance(top_SR.point, sr.point)
                        time_passed = int(time.time() - msg_s_time)

                        srp =  f'{sr.id} {sr.smooth_score:{pf}} [{sr.score:{pf}} {difs}] {top_SR.id}:{dist_to_max:.3f}'
                        srp += f'  avg/mom:{avg_dst:.3f}/{mom_dst:.3f}  {time_passed}s'
                        if new_sampling_config: srp += f'  new sampling config: {sampling_config}'
                        print(srp)

                        # new MAX report
                        if gots_new_max:

                            msr = f'_newMAX: {top_SR.id} {top_SR.smooth_score:{pf}} [{top_SR.score:{pf}}] {point_str(top_SR.point)}\n'

                            prev_sr = srl.get_SR(prev_max_sr_id)
                            dp = srl.paspa.distance(prev_sr.point, top_SR.point) if prev_sr else 0

                            msr += f' dst_prev:{dp:.3f}\n'
                            for nps in NP_SMOOTH:
                                ss_np, avd, all_sc = srl.smooth_point(top_SR, nps)
                                msr += f'  NPS:{nps} {ss_np:{pf}} [{max(all_sc):{pf}}-{min(all_sc):{pf}}] {avd:.3f}\n'
                            print(msr)

                        if top_show_freq and len(srl) % top_show_freq == 0:
                            speed = int((time.time()-top_time) / top_show_freq)
                            top_time = time.time()
                            top_speed_save.append(speed)
                            if len(top_speed_save) > 10: top_speed_save.pop(0)
                            diff = int(speed - (sum(top_speed_save) / len(top_speed_save)))
                            print(f'\n ### hpmser speed: {speed} sec/task, diff: {"+" if diff >=0 else "-"}{abs(diff)} sec')

                            print(srl.nice_str(n_top=4, top_nps=NP_SMOOTH, all_nps=None))

                    if tbwr:
                        scores_all.append(sr.score)
                        score_diff = sr.score - msg_est_score
                        score_avg = sum(scores_all)/len(scores_all)
                        step = len(srl)
                        tbwr.add(avg_dst,         'hpmser/avg_dst',                    step)
                        tbwr.add(score_avg,       'hpmser/score_avg',                  step)
                        tbwr.add(sr.score,        'hpmser/score_current',              step)
                        tbwr.add(score_diff,      'hpmser/space_estimation_error',     step)
                        tbwr.add(abs(score_diff), 'hpmser/space_estimation_error_abs', step)

                    if len(srl) == n_loops:
                        if verb>0: print(f'{n_loops} loops done!')
                        break

    except KeyboardInterrupt:
        if verb>0: print(' > hpmser_GX KeyboardInterrupt-ed..')
        raise KeyboardInterrupt # raise exception for OMPRunner

    except Exception as e:
        print(f'hpmser_GX Exception: {str(e)}')
        raise e

    finally:
        omp.exit()

        srl.save(folder=f'{hpmser_FD}/{name}')

        results_str = srl.nice_str(top_nps=NP_SMOOTH)
        if hpmser_FD:
            with open( f'{hpmser_FD}/{name}/{name}_results.txt', 'w') as file: file.write(results_str)
        if verb>0: print(f'\n{results_str}')

        return srl


"""

 2022 (c) piteren

    Envy is a base environment interface

    RLEnvy is an interface that defines base RL methods used by Actor or Trainer (..or other objects)

        RLEnvy returns reward after step (run). This value may (should?) be processed / overridden by Trainer.
        Trainer is supposed to train Actor using information of reward that he defines / corrects observing an Envy.
        He may apply discount, factor, moving average etc to those values.
        (Actor does not need reward to act with policy.)

"""

from abc import abstractmethod, ABC
import numpy as np
from typing import List, Tuple, Optional

from pypaq.lipytools.pylogger import get_pylogger
from pypaq.R4C.helpers import RLException



# base Environment interface
class Envy(ABC):

    def __init__(
            self,
            seed: int,
            logger=     None,
            loglevel=   20):
        self._log = logger or get_pylogger(level=loglevel)
        self.seed = seed
        self._log.info(f'*** Envy *** initialized')
        self._log.info(f'> seed:      {self.seed}')
        self._log.info(f'> max steps: {self.get_max_steps()}')

    # plays action, goes to new state
    @abstractmethod
    def run(self, action: object) -> None: pass

    # returns observation of current state
    @abstractmethod
    def get_observation(self) -> object: pass

    # resets Envy (self) to initial state with given seed
    @abstractmethod
    def _reset_with_seed(self, seed: int): pass

    # resets Envy (self) to initial state
    def reset(self):
        self._reset_with_seed(seed=self.seed)
        self.seed += 1

    # returns max number of steps in one episode, None means infinite
    @abstractmethod
    def get_max_steps(self) -> Optional[int]: pass


# adds to Envy methods needed by base RL algorithms
class RLEnvy(Envy, ABC):

    def run(self, action: object) -> Tuple[
        float,  # reward
        bool,   # is terminal
        bool    # has won
    ]: pass

    # Envy current state rendering (for debug, preview etc.)
    def render(self): pass

    # prepares numpy vector from observation, it may be implemented by RLEnvy, but is not mandatory, otherwise Actor should implement on itself
    def prep_observation_vec(self, observation: object) -> np.ndarray:
        raise RLException('RLEnvy not implemented prep_observation_vec()')


# interface of RL Environment with finite actions number
class FiniteActionsRLEnvy(RLEnvy):

    def __init__(self, **kwargs):
        RLEnvy.__init__(self, **kwargs)
        self._log.info(f'*** FiniteActionsRLEnvy *** initialized')
        self._log.info(f'> num_actions: {self.num_actions()}')

    # returns list of valid actions
    @abstractmethod
    def get_valid_actions(self) -> List[object]: pass

    # returns number of Envy actions
    def num_actions(self) -> int:
        return len(self.get_valid_actions())


"""

 2022 (c) piteren

    NN TF graph elements and tools

"""

import tensorflow as tf
from typing import List



# default initializer for variables of graph (recommended by TF, BERT: stddev = 0.02)
def my_initializer(seed=12321, stddev=0.02):
    # TODO: consider GlorotUniform (Keras default) - compare on some cases
    return tf.keras.initializers.TruncatedNormal(stddev=stddev, seed=seed)

# GeLU (Gaussian Error Linear Unit) activation https://arxiv.org/abs/1606.08415
def gelu(x):
    # ga = x * 0.5 * (1.0 + tf.tanh((np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3))))) # old version from TF.1.X
    return tf.keras.activations.gelu(x)

# replaces nan values @tensor with zero
def replace_nan_with_zero(tensor): return tf.where(tf.math.is_nan(tensor), tf.zeros_like(tensor), tensor)

# scales learning rate with warmUp and annealing (after warmUp)
def lr_scaler(
        baseLR,                            # initial learning rate
        g_step: tf.Tensor,              # global step
        warm_up: int or None=   1000,   # warmup steps, None or 0 turns-off
        ann_base: float=        0.999,  # annealing base, None or 1 for turn-off
        ann_step: float=        1.0,    # annealing step, higher value speeds up annealing
        n_wup_off: float=       2.0,    # N warmUp offset of annealing
        verb=                   0):

    if verb>0: print(f'\nbuilding lR scaling graph for baseLR: {baseLR} ..')
    g_step_fl = tf.cast(g_step, dtype=tf.float32)
    if warm_up is None: warm_up = 0
    lR = baseLR

    if warm_up:
        ratioWm = tf.reduce_min([g_step_fl, warm_up]) / warm_up # warmUp ratio
        lR = baseLR * ratioWm # learning rate with warmup
        if verb>0: print(f'applied warmUp ({warm_up}) to lR')
    if ann_base is not None and ann_base != 1:
        gStep_offs = tf.reduce_max([0, g_step_fl - warm_up * n_wup_off]) # offset by warmUpSteps
        lR *= ann_base ** (gStep_offs * ann_step) # learning rate with annealing
        if verb>0: print(f'applied annealing to lR ({ann_base:.5f},{ann_step:.5f})')
    return lR

# gradient clipping layer (by clip_value or AVT algorithm)
def grad_clipper_AVT(
        variables: List[tf.Variable],   # list of variables
        gradients: List[tf.Tensor],     # list of gradients (for variables)
        ggnorm_avt: tf.Variable,        # variable that holds averaged_over_time global_gradients_norm
        optimizer: tf.keras.optimizers.Optimizer,
        clip_value=         None,       # clipping value, for None clips with AVT
        avt_window: int=    100,        # width of averaging window (number of steps)
        avt_max_upd: float= 1.5,        # single step max factor of avt update
        do_clip=            True,       # disables clipping (just GN calculations)
        verb=               0):

    if verb > 0: print(f'grad_clipper_AVT: start value {ggnorm_avt}, avt_window {avt_window}, avt_max_upd {avt_max_upd:.1f}')

    ggnorm = tf.linalg.global_norm(gradients) # global norm of gradients

    avt_update = tf.reduce_min([ggnorm, avt_max_upd * ggnorm_avt]) # single value to update AVTG with (current GNorm or clipped to max value)

    # new value
    new_val = (ggnorm_avt * (avt_window - 1) + avt_update) / avt_window
    ggnorm_avt.assign(new_val)

    if do_clip:
        gradients, _ = tf.clip_by_global_norm(
            t_list=     gradients,
            clip_norm=  clip_value or ggnorm_avt,
            use_norm=   ggnorm)
        if verb>0: print(f' >> is clipping gradients {"with value" if clip_value else "with AVT"}')
    elif verb>0: print(' >> not doing clipping')

    optimizer.apply_gradients(grads_and_vars=zip(gradients,variables))

    return {
        'gradients':    gradients,
        'ggnorm':       ggnorm}


"""

 2020 (c) piteren

    NN Batcher
        takes data and prepares batches
        data for training, validation or testing is a dict: {key: np.array or torch.tensor}
        batch is prepared from each key

"""

import numpy as np
from typing import Dict, Optional, Tuple

from pypaq.lipytools.pylogger import get_pylogger


BATCHING_TYPES = [
    'base',         # prepares batches in order of given data
    'random',       # basic random sampling
    'random_cov']   # random sampling with full coverage of data


class BatcherException(Exception):
    pass


class Batcher:

    def __init__(
            self,
            data_TR: Dict[str,np.ndarray],
            data_VL: Optional[Dict[str,np.ndarray]]=    None,
            data_TS: Optional[Dict[str,np.ndarray]]=    None,
            batch_size: int=                            16,
            bs_mul: int=                                2,      # VL & TS batch_size multiplier
            batching_type: str=                         'random_cov',
            seed=                                       123,
            logger=                                     None):

        self.__log = logger or get_pylogger()

        self.seed_counter = seed

        if batching_type not in BATCHING_TYPES:
            raise BatcherException('ERR: unknown batching_type!')

        self.btype = batching_type

        self._data_keys = sorted(list(data_TR.keys()))

        self._data_TR = data_TR
        self._data_VL = data_VL
        self._data_TS = data_TS
        self._data_len_TR = self._data_TR[self._data_keys[0]].shape[0]

        self._batch_size = None
        self.set_batch_size(batch_size)
        self._bs_mul = bs_mul

        self._data_ixmap = []

        self._VL_batches = None
        self._TS_batches = None

        self.__log.info(f'*** Batcher *** initialized with {self._data_len_TR} samples of data in keys, batch size: {batch_size}')
        self.__log.debug('> Batcher keys:')
        for k in self._data_keys:
            self.__log.debug(f'>> {k}, shape: {self._data_TR[k].shape}, type:{type(self._data_TR[k][0])}')

    def _extend_ixmap(self):

        if self.btype == 'base':
            self._data_ixmap += list(range(self._data_len_TR))

        if self.btype == 'random':
            self._data_ixmap += np.random.choice(
                a=          self._data_len_TR,
                size=       self._batch_size,
                replace=    False).tolist()

        if self.btype == 'random_cov':
            new_ixmap = np.arange(self._data_len_TR)
            np.random.shuffle(new_ixmap)
            new_ixmap = new_ixmap.tolist()
            self._data_ixmap += new_ixmap

    def set_batch_size(self, bs: int):
        if bs > self._data_len_TR:
            raise BatcherException('ERR: cannot set batch size larger than given TR data!')
        self._batch_size = bs

    def get_batch(self) -> dict:

        # set seed
        np.random.seed(self.seed_counter)
        self.seed_counter += 1

        if len(self._data_ixmap) < self._batch_size: self._extend_ixmap()
        
        indexes = self._data_ixmap[:self._batch_size]
        self._data_ixmap = self._data_ixmap[self._batch_size:]
        return {k: self._data_TR[k][indexes] for k in self._data_keys}

    @staticmethod
    def __split_data(data:dict, size:int) -> list:
        split = []
        counter = 0
        keys = list(data.keys())
        while counter*size < len(data[keys[0]]):
            split.append({k: data[k][counter*size:(counter+1)*size] for k in keys})
            counter += 1
        return split

    def get_VL_batches(self) -> list:
        if self._VL_batches is None:
            if self._data_VL is None:
                raise BatcherException('ERR: cannot prepare VL batches - data nat given')
            self._VL_batches =  Batcher.__split_data(self._data_VL, self._batch_size * self._bs_mul)
        return self._VL_batches

    def get_TS_batches(self) -> list:
        if self._TS_batches is None:
            if self._data_TS is None:
                err = 'ERR: cannot prepare TS batches - data nat given'
                self.__log.error(err)
                raise BatcherException(err)
            self._TS_batches = Batcher.__split_data(self._data_TS, self._batch_size * self._bs_mul)
        return self._TS_batches

    def get_data_size(self) -> Tuple[int,int,int]:
        k = self._data_keys[0]
        return self._data_TR[k].shape[0], self._data_VL[k].shape[0] if self._data_VL else 0, self._data_TS[k].shape[0] if self._data_TS else 0


def split_data_TR(
        data: Dict[str,np.ndarray],
        split_VL: float=    0.0,  # if > 0.0 and not data_VL then factor of data_TR will be put to data_VL
        split_TS: float=    0.0,  # if > 0.0 and not data_TS then factor of data_TR will be put to data_TS
        seed=               123,
) -> Tuple[Dict[str,np.ndarray], Optional[Dict[str,np.ndarray]], Optional[Dict[str,np.ndarray]]]:

    if split_VL == 0.0 and split_TS == 0.0:
        return data, None, None
    else:
        data_keys = list(data.keys())
        d_len = data[data_keys[0]].shape[0]
        np.random.seed(seed)
        indices = np.random.permutation(d_len)
        nVL = int(d_len * split_VL)
        nTS = int(d_len * split_TS)
        nTR = d_len-nVL-nTS
        indices_TR = indices[:nTR]
        indices_VL = indices[nTR:nTR+nVL] if nVL else None
        indices_TS = indices[nTR+nVL:] if nTS else None

        dTR = {}
        dVL = {}
        dTS = {}
        for k in data_keys:
            dTR[k] = data[k][indices_TR]
            if indices_VL is not None:
                dVL[k] = data[k][indices_VL]
            if indices_TS is not None: dTS[k] = data[k][indices_TS]

        return dTR, dVL or None, dTS or None


from typing import Optional, Tuple
import torch

from pypaq.torchness.types import ACT, INI, TNS, DTNS
from pypaq.torchness.base_elements import bert_initializer, my_initializer, TorchnessException
from pypaq.torchness.layers import LayDense, TF_Dropout, LayConv1D, LayRES, zeroes


# Block (Layer) of EncDRT (LN > Dense or two_with_drop_in_between > drop > RES with drop)
class LayBlockDRT(torch.nn.Module):

    def __init__(
            self,
            in_width: int,
            do_scaled_dns: bool=        False,          # two denses (True) or single dense (False)
            dns_scale: int=             4,              # up-scale for first dense of two
            activation: ACT=            torch.nn.ReLU,
            interlay_dropout: float=    0.0,            # dropout in between two denses
            lay_dropout: float=         0.0,            # dropout after dense/s
            residual: bool=             True,           # residual yes/no
            res_dropout: float=         0.0,            # dropout on residual connection
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(LayBlockDRT, self).__init__()

        if initializer is None: initializer = my_initializer

        self.ln_in = torch.nn.LayerNorm(
            normalized_shape=   in_width,
            device=             device,
            dtype=              dtype)

        self.denses = []
        if do_scaled_dns:
            # dense (scale up) with activation
            self.denses.append(LayDense(
                in_features=    in_width,
                out_features=   in_width * dns_scale,
                activation=     activation,
                bias=           True,
                device=         device,
                dtype=          dtype,
                initializer=    initializer))
            # dense (scale down) without activation
            self.denses.append(LayDense(
                in_features=    in_width * dns_scale,
                out_features=   in_width,
                activation=     None,
                bias=           True,
                device=         device,
                dtype=          dtype,
                initializer=    initializer))
        else:
            # just single dense, with activation
            self.denses.append(LayDense(
                in_features=    in_width,
                out_features=   in_width,
                activation=     activation,
                bias=           True,
                device=         device,
                dtype=          dtype,
                initializer=    initializer))
        for dix, l in enumerate(self.denses): self.add_module(f'dense{dix}', l)

        self.drop_interlay = torch.nn.Dropout(p=interlay_dropout) if interlay_dropout and do_scaled_dns else None

        self.drop_lay = torch.nn.Dropout(p=lay_dropout) if lay_dropout else None

        self.res = LayRES(in_features=in_width, dropout=res_dropout) if residual else None

    def forward(self, input:TNS) -> DTNS:

        out = self.ln_in(input)

        out = self.denses[0](out)
        zsL = [zeroes(out)]

        if len(self.denses) > 1: # there is second one, without activation

            if self.drop_interlay:
                out = self.drop_interlay(out)

            out = self.denses[1](out)

        if self.drop_lay:
            out = self.drop_lay(out)

        if self.res:
            out = self.res(input=out, bypass=input)

        return {
            'out':  out,
            'zsL':  zsL}


# Deep Residual encoder based on stacked LayBlockDRT
class EncDRT(torch.nn.Module):

    def __init__(
            self,
            in_width: int,
            in_dropout: float=          0.0,            # dropout on input
            n_layers: int=              6,
            shared_lays: bool=          False,          # shared variables in enc_layers
            lay_width: Optional[int]=   None,           # for None matches input width
            do_scaled_dns: bool=        True,
            dns_scale: int=             4,              # scale(*) of first dense
            activation: ACT=            torch.nn.ReLU,  # gelu is really worth a try
            lay_dropout: float=         0.0,            # dropout after two denses
            residual: bool=             True,           # residual yes/no
            res_dropout: float=         0.0,            # dropout on residual connection
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(EncDRT, self).__init__()

        if initializer is None: initializer = my_initializer

        self.in_drop_lay = torch.nn.Dropout(p=in_dropout) if in_dropout else None

        self.in_width = in_width
        self.lay_width = lay_width or self.in_width
        self.projection_lay = LayDense(
            in_features=    self.in_width,
            out_features=   self.lay_width,
            activation=     None,
            bias=           False,
            device=         device,
            dtype=          dtype,
            initializer=    initializer) if self.lay_width != self.in_width else None

        self.ln_in = torch.nn.LayerNorm(
            normalized_shape=   self.lay_width,
            device=             device,
            dtype=              dtype)

        num_layers_to_build = 1 if shared_lays else n_layers
        self.drt_lays = [LayBlockDRT(
            in_width=       self.lay_width,
            do_scaled_dns=  do_scaled_dns,
            dns_scale=      dns_scale,
            activation=     activation,
            lay_dropout=    lay_dropout,
            residual=       residual,
            res_dropout=    res_dropout,
            device=         device,
            dtype=          dtype,
            initializer=    initializer
        ) for _ in range(num_layers_to_build)]
        for lix,lay in enumerate(self.drt_lays): self.add_module(f'lay_drt_{lix}',lay)
        if shared_lays and n_layers > 1: self.drt_lays *= n_layers

    def forward(self, input:TNS) -> DTNS:

        zsL = []

        out = input

        if self.in_drop_lay: # input dropout
            out = self.in_drop_lay(out)

        if self.projection_lay: # input projection, no activation <- do not catch zeroes
            out = self.projection_lay(out)

        out = self.ln_in(out)

        for drt_lay in self.drt_lays:
            lay_out = drt_lay(out)
            out = lay_out['out']
            zsL += lay_out['zsL']

        return {
            'out':  out,
            'zsL':  zsL}


# Block (Layer) of EncCNN (LN > CNN > act > drop > RES > LayBlockDRT), number of parameters: kernel*in_features*n_filters
class LayBlockCNN(torch.nn.Module):

    def __init__(
            self,
            n_filters: int,                             # num of filters
            kernel_size: int=           3,              # layer kernel
            activation: ACT=            torch.nn.ReLU,  # global enc activation func
            lay_dropout: float=         0.0,
            res_dropout: float=         0.0,
            # lay_DRT
            do_ldrt=                    False,          # lay DRT - build or not
            ldrt_do_scaled_dns: bool=   True,
            ldrt_dns_scale: int=        4,
            ldrt_drop: float or None=   0.0,
            ldrt_residual: bool=        True,
            ldrt_res_dropout: float=    0.0,
            # other
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(LayBlockCNN, self).__init__()

        if kernel_size % 2 != 1: raise TorchnessException('LayBlockCNN kernel_size cannot be even number')
        self.kernel_size = kernel_size

        self.lay_ln = torch.nn.LayerNorm(
            normalized_shape=   n_filters,
            device=             device,
            dtype=              dtype)

        self.lay_conv1D = LayConv1D(
            in_features=    n_filters,
            n_filters=      n_filters,
            kernel_size=    self.kernel_size,
            padding=        'valid',
            device=         device,
            dtype=          dtype,
            activation=     None,
            initializer=    initializer)

        self.activation = activation() if activation else None

        self.lay_drop = torch.nn.Dropout(p=lay_dropout) if lay_dropout else None

        self.res = LayRES(in_features=n_filters, dropout=res_dropout)

        self.lay_DRT = LayBlockDRT(
            in_width=       n_filters,
            do_scaled_dns=  ldrt_do_scaled_dns,
            dns_scale=      ldrt_dns_scale,
            activation=     activation,
            lay_dropout=    ldrt_drop,
            residual=       ldrt_residual,
            res_dropout=    ldrt_res_dropout,
            device=         device,
            dtype=          dtype,
            initializer=    initializer) if do_ldrt else None

    def forward(self, input:TNS, history:Optional[TNS]=None) -> DTNS:

        zsL = []

        out = self.lay_ln(input)

        if history is None:
            in_sh = list(input.shape)
            pad_width = int((self.kernel_size-1)/2)
            in_sh[-2] = pad_width
            pad = torch.zeros(in_sh)
            conc = [pad, out, pad] # pad both sides (encoder)
        else:
            conc = [history, out] # concatenate with history (casual encoder)
        out = torch.concat(conc, dim=-2)

        out = self.lay_conv1D(out)

        if self.activation:
            out = self.activation(out)
            zsL.append(zeroes(out))

        if self.lay_drop:
            out = self.lay_drop(out)

        out = self.res(input=out, bypass=input)

        if self.lay_DRT:
            lay_out = self.lay_DRT(out)
            out = lay_out['out']
            zsL += lay_out['zsL']

        return {
            'out':      out,
            'state':    torch.split(input, split_size_or_sections=self.kernel_size-1, dim=-2)[-1],
            'zsL':      zsL}


# CNN 1D Encoder (for sequences), number of parameters: projection + n_layers*LayBlockCNN
class EncCNN(torch.nn.Module):

    def __init__(
            self,
            in_features: int,                           # input num of channels
            time_drop: float=           0.0,
            feat_drop: float=           0.0,
            # layer
            shared_lays: bool=          False,          # shared variables in enc_layers
            n_layers :int=              6,              # num of layers
            kernel_size :int=           3,              # layer kernel
            n_filters :Optional[int]=   None,           # num of filters
            activation: ACT=            torch.nn.ReLU,  # global enc activation func
            lay_dropout: float=         0.0,
            res_dropout: float=         0.0,
            # lay_DRT
            do_ldrt=                    False,          # lay DRT - build or not
            ldrt_do_scaled_dns: bool=   True,
            ldrt_dns_scale: int=        4,
            ldrt_drop: float or None=   0.0,
            ldrt_residual: bool=        True,
            ldrt_res_dropout: float=    0.0,
            # other
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(EncCNN, self).__init__()

        self.in_features = in_features
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.n_filters = n_filters or self.in_features

        self.in_TFdrop_lay = TF_Dropout(
            time_drop=  time_drop,
            feat_drop=  feat_drop) if time_drop or feat_drop else None

        self.projection_lay = LayDense(
            in_features=    self.in_features,
            out_features=   self.n_filters,
            activation=     None,
            bias=           False,
            device=         device,
            dtype=          dtype,
            initializer=    initializer) if self.in_features != self.n_filters else None

        num_blocks_to_build = 1 if shared_lays else self.n_layers

        self.blocks = [LayBlockCNN(
            n_filters=          self.n_filters,
            kernel_size=        self.kernel_size,
            activation=         activation,
            lay_dropout=        lay_dropout,
            res_dropout=        res_dropout,
            do_ldrt=            do_ldrt,
            ldrt_do_scaled_dns= ldrt_do_scaled_dns,
            ldrt_dns_scale=     ldrt_dns_scale,
            ldrt_drop=          ldrt_drop,
            ldrt_residual=      ldrt_residual,
            ldrt_res_dropout=   ldrt_res_dropout,
            device=             device,
            dtype=              dtype,
            initializer=        initializer) for _ in range(num_blocks_to_build)]

        for bix,block in enumerate(self.blocks): self.add_module(f'block_{bix}',block)

        if shared_lays and self.n_layers > 1: self.blocks *= self.n_layers

        self.out_ln = torch.nn.LayerNorm(
            normalized_shape=   self.n_filters,
            device=             device,
            dtype=              dtype)

    # prepares initial history for casual mode, history has shape [.., n_layers, kernel_size-1, n_filters]
    def get_zero_history(self, input:TNS) -> TNS:
        in_sh = list(input.shape)
        in_sh.insert(-2, self.n_layers)
        in_sh[-2] = self.kernel_size - 1
        in_sh[-1] = self.n_filters
        return torch.zeros(in_sh)

    def forward(self, input:TNS, history:Optional[TNS]=None) -> DTNS:

        states = []  # here we will store block states to concatenate them finally
        zsL = []

        if self.in_TFdrop_lay:
            input = self.in_TFdrop_lay(input)

        if self.projection_lay:
            input = self.projection_lay(input)

        output = input  # for 0 layers case
        histories = torch.split(history,1,dim=-3) if history is not None else [None]*self.n_layers
        for block,hist in zip(self.blocks, histories):
            if hist is not None: hist = torch.squeeze(hist, dim=-3)
            block_out = block(output, history=hist)
            output = block_out['out']
            states.append(torch.unsqueeze(block_out['state'], dim=-3))
            zsL += block_out['zsL']

        output = self.out_ln(output)

        return {
            'out':      output,
            'state':    torch.concat(states,dim=-3),
            'zsL':      zsL}


# QKV_linear_projection + QKV_scaled_dot_product_attention + linear_out_projection
class MyMHA(torch.nn.MultiheadAttention):

    # replaces xavier with bert_initializer
    def _reset_parameters(self):

        if self._qkv_same_embed_dim:
            bert_initializer(self.in_proj_weight)
        else:
            bert_initializer(self.q_proj_weight)
            bert_initializer(self.k_proj_weight)
            bert_initializer(self.v_proj_weight)
        bert_initializer(self.out_proj.weight)

        if self.in_proj_bias is not None:
            torch.nn.init.zeros_(self.in_proj_bias)
            torch.nn.init.zeros_(self.out_proj.bias)
        if self.bias_k is not None:
            bert_initializer(self.bias_k)
        if self.bias_v is not None:
            bert_initializer(self.bias_v)


# Block (Layer) of EncTNS (based on torch.nn.modules.transformer.TransformerEncoderLayer)
class LayBlockTNS(torch.nn.Module):

    def __init__(
            self,
            d_model: int=       512,
            nhead: int=         8,
            dns_scale: int=     4,              # up-scale for first dense of two
            dropout: float=     0.1,
            dropout_att: float= 0.0,            # in original (torch.nn..) implementation dropout_att == dropout
            activation: ACT=    torch.nn.ReLU,
            dropout_res: float= 0.0,            # dropout on residual bypass
            device=             None,
            dtype=              None):

        super(LayBlockTNS, self).__init__()

        self.norm1 = torch.nn.LayerNorm(
            normalized_shape=   d_model,
            device=             device,
            dtype=              dtype)

        self.self_attn = MyMHA(
            embed_dim=      d_model,
            num_heads=      nhead,
            dropout=        dropout_att,
            bias=           True,
            add_bias_kv=    False,
            add_zero_attn=  False,
            kdim=           None,
            vdim=           None,
            batch_first=    True,
            device=         device,
            dtype=          dtype)

        self.dropout1 = torch.nn.Dropout(p=dropout) if dropout else None

        self.res1 = LayRES(in_features=d_model, dropout=dropout_res)

        # (LN > Dense or two_with_drop_in_between > drop > RES with drop)
        self.lay_drt = LayBlockDRT(
            in_width=           d_model,
            do_scaled_dns=      True,
            dns_scale=          dns_scale,
            activation=         activation,
            interlay_dropout=   dropout, # INFO: TF implementation has not this dropout
            lay_dropout=        dropout,
            residual=           True,
            res_dropout=        dropout_res,
            device=             device,
            dtype=              dtype,
            initializer=        bert_initializer)

    def forward(
            self,
            src: TNS,
            task_query: Optional[TNS]=              None,           # forces task-attention mode (TAT)
            src_mask: Optional[TNS]=                None,
            src_key_padding_mask: Optional[TNS]=    None) -> DTNS:

        x = src

        if task_query is None: x = self.norm1(x) # norm first https://arxiv.org/pdf/2002.04745v1.pdf
        else: task_query = self.norm1(task_query) # LN on task_query

        x = self.self_attn(
            query=              x if task_query is None else task_query,
            key=                x,
            value=              x,
            key_padding_mask=   src_key_padding_mask,
            need_weights=       False,
            attn_mask=          src_mask)[0]

        if self.dropout1:
            x = self.dropout1(x)

        bypass = src if task_query is None else task_query
        x = self.res1(input=x, bypass=bypass)

        return self.lay_drt(x)


# Transformer Encoder (based on torch.nn.modules.transformer.TransformerEncoder)
class EncTNS(torch.nn.Module):

    def __init__(
            self,
            num_layers: int=                        6,
            num_layers_TAT: int=                    0,
            shared_lays: Optional[Tuple[int,...]]=  None,   # tuple defines layers groups with shared variables, e.g.: (2,2,2)
            max_seq_len: Optional[int]=             None,   # when given (int) adds positional embeddings (PE) to seq
            # block params
            d_model: int=                           512,
            nhead: int=                             8,
            dns_scale: int=                         4,
            dropout: float=                         0.1,
            dropout_att: float=                     0.0,
            activation: ACT=                        torch.nn.ReLU,
            dropout_res: float=                     0.0,
            device=                                 None,
            dtype=                                  None):

        super(EncTNS, self).__init__()

        # positional embeddings (trainable)
        self.pos_emb = None
        if max_seq_len:
            self.pos_emb = torch.nn.Parameter(torch.empty((max_seq_len, d_model), device=device, dtype=dtype))
            bert_initializer(self.pos_emb)

        # manage layers number
        num_layers_to_build = num_layers + num_layers_TAT
        if shared_lays is not None:

            if not sum(shared_lays) == num_layers_to_build:
                raise TorchnessException('Sum of shared layers must be equal to total num of layers (num_layers + num_layers_TAT)')

            num_layers_to_build = len(shared_lays)

        layers = [LayBlockTNS(
            d_model=        d_model,
            nhead=          nhead,
            dns_scale=      dns_scale,
            dropout=        dropout,
            dropout_att=    dropout_att,
            activation=     activation,
            dropout_res=    dropout_res,
            device=         device,
            dtype=          dtype,
        ) for _ in range(num_layers_to_build)]
        for lix,lay in enumerate(layers): self.add_module(f'lay_{lix}',lay)

        if shared_lays is not None:
            exp_layers = []
            for lay,mul in zip(layers,shared_lays):
                exp_layers += [lay]*mul
            layers = exp_layers

        self.layers = layers[:num_layers]
        self.layers_TAT = layers[num_layers:]

        self.norm = torch.nn.LayerNorm(
            normalized_shape=   d_model,
            device=             device,
            dtype=              dtype)

    def forward(self, src:TNS, mask:Optional[TNS]=None) -> DTNS:

        output = src

        # add positional embeddings if needed
        if self.pos_emb is not None:
            output += self.pos_emb[:output.size(-2)]

        zsL = []
        for mod in self.layers:
            block_out = mod(src=output, src_mask=mask)
            output = block_out['out']
            zsL += block_out['zsL']

        # pass through task-attention layers
        if self.layers_TAT:
            seq = output
            task_query = torch.mean(seq, dim=-2, keepdim=True) # first initial reduce
            for mod in self.layers_TAT:
                block_out = mod(src=seq, task_query=task_query, src_mask=mask)
                task_query = block_out['out']
                zsL += block_out['zsL']
            output = task_query

        output = self.norm(output)

        return {
            'out':  output,
            'zsL':  zsL}


"""

 2018 (c) piteren

 MultiSaver for NN models
 - implements support for many savers
 - stores (saves and loads) variables in subfolders named with keys of 'vars' dict (default 'ALL')
 - supports list of savers (names for different savers for same variable list)

"""

from typing import Dict, List, Optional, Tuple

from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.pylogger import get_pylogger
from pypaq.neuralmess.get_tf import tf


class MultiSaver:

    def __init__(
            self,
            model_name :str,
            vars: List or Dict[str,List],               # variables may be given in list or dict {key: list} where key is name for the list
            save_TFD :str,                              # save topfolder
            savers: Tuple or List=          (None,),    # tuple of savers names managed by MultiSaver
            max_keep: Optional[List[int]]=  None,       # for each saver keeps N last ckpt saves, for None: 1 for each saver
            session=                        None,
            logger=                         None):
            
        self._log = logger or get_pylogger()

        self.save_FD = f'{save_TFD}/{model_name}'
        self.model_name = model_name

        if type(vars) is list: vars = {'ALL': vars}
        self.vars = vars

        self.session = session

        if max_keep is None: max_keep = [1 for _ in savers]
        assert len(max_keep) == len(savers)

        self.savers = {}
        for ix in range(len(savers)):
            self.savers[savers[ix]] = {}
            for var in self.vars:
                self.savers[savers[ix]][var] = tf.train.Saver(
                    var_list=               self.vars[var],
                    pad_step_number=        True,
                    save_relative_paths=    True,
                    max_to_keep=            max_keep[ix])

        self.s_step = {sv: {var: 0 for var in self.vars} for sv in savers} # self step per saver per vars

        self._log.info(f'*** MultiSaver for {self.model_name} model')
        self._log.debug(f'> MultiSaver folder: {self.save_FD}')
        self._log.debug(f'> gots {len(self.vars)} lists of variables')
        for var in self.vars: self._log.log(5, f' >> {var} - {len(self.vars[var])} variables')
        self._log.debug(f'> for every var list gots {len(savers)} savers: {savers}')

    # saves checkpoint of given saver
    def save(
            self,
            saver=      None,   # saver name
            step :int=  None,   # for None uses self step
            session=    None):

        assert saver in self.savers, 'ERR: unknown saver'

        prep_folder(self.save_FD)
        sv_name = f' {saver}' if saver else ''
        self._log.debug(f'MultiSaver{sv_name} saves variables...')

        for var in self.vars:
            ckpt_path = f'{self.save_FD}/{var}/{self.model_name}'
            if saver: ckpt_path += f'_{saver}'
            ckpt_path += '.ckpt'

            if not session: session = self.session

            latest_filename = 'checkpoint'
            if saver: latest_filename += f'_{saver}'

            self.savers[saver][var].save(
                sess=               session,
                save_path=          ckpt_path,
                global_step=        step if step else self.s_step[saver][var],
                latest_filename=    latest_filename,
                write_meta_graph=   False,
                write_state=        True)
            self.s_step[saver][var] += 1
            self._log.debug(f' > saved variables {var}')

    # loads last checkpoint of given saver
    def load(
            self,
            saver=                          None,
            session :tf.compat.v1.Session=  None,
            allow_init=                     True):

        if not session: session = self.session

        for var in self.vars:
            # look for checkpoint
            latest_filename = 'checkpoint'
            if saver: latest_filename += '_' + saver
            ckpt = tf.train.latest_checkpoint(
                checkpoint_dir=     self.save_FD + '/' + var,
                latest_filename=    latest_filename) if self.save_FD else None

            if ckpt:
                """
                from tensorflow.python.tools.inspect_checkpoint import print_tensors_in_checkpoint_file
                print_tensors_in_checkpoint_file(
                    file_name=      ckpt,
                    tensor_name=    '',
                    all_tensors=    False)
                """
                self.savers[saver][var].restore(session, ckpt)
                self._log.debug(f'Variables {var} restored from checkpoint {saver if saver else ""}')

            else:
                assert allow_init, 'Err: saver load failed: checkpoint not found and not allowInit'
                session.run(tf.initializers.variables(self.vars[var]))
                self._log.debug(f'No checkpoint found, variables {var} initialized with default initializer')


import csv
import json
import os
import pickle
import shutil
import yaml
from pathlib import Path
from typing import Union, Dict, List


# ********************************************* for raise_exception=False each reader will return None if file not found

# pickle read
def r_pickle(
        file_path,
        obj_type=           None, # if obj_type is given checks for compatibility with given type
        raise_exception=    False):
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None

    # obj = pickle.load(open(file_path, 'rb')) << replaced by:
    with open(file_path, 'rb') as file: obj = pickle.load(file)

    if obj_type: assert type(obj) is obj_type, f'ERROR: obj from file is not {str(obj_type)} type !!!'
    return obj

# pickle write
def w_pickle(
        obj,
        file_path):
    with open(file_path, 'wb') as file:
        pickle.dump(obj, file)

# json read
def r_json(
        file_path,
        raise_exception=    False):
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return json.load(file)

# json write
def w_json(
        data: Union[Dict,List],
        file_path):
    with open(file_path, 'w', encoding='utf-8') as file:
        json.dump(data, file, indent=4, ensure_ascii=False)

# jsonl read
def r_jsonl(
        file_path,
        raise_exception=False):
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return [json.loads(line) for line in file]

# jsonl write
def w_jsonl(
        data: List[dict],
        file_path):
    with open(file_path, 'w', encoding='utf-8') as file:
        for d in data:
            json.dump(d, file, ensure_ascii=False)
            file.write('\n')

# csv read
def r_csv(
        file_path,
        raise_exception=    False):
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, newline='') as f:
        reader = csv.reader(f)
        return [row for row in reader][1:]

# yaml read
def r_yaml(
        file_path,
        raise_exception=    False):
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path) as file:
        return yaml.load(file, yaml.Loader)


def get_dir(path: Union[str, Path]):
    path = str(path)
    path_split = path.split('/')
    if path_split[-1].find('.') != -1: path_split = path_split[:-1]
    return '/'.join(path_split)


def prep_folder(
        path: Union[str, Path],  # folder or file path
        flush_non_empty=    False):
    folder_path = get_dir(path)
    if flush_non_empty and os.path.isdir(folder_path): shutil.rmtree(folder_path)
    os.makedirs(folder_path, exist_ok=True)


def list_dir(path: Union[str, Path]) -> Dict[str,List]:
    ls = os.listdir(path)
    lsD = {'files':[], 'dirs':[]}
    for e in ls:
        lsD['files' if os.path.isfile(f'{path}/{e}') else 'dirs'].append(e)
    return lsD



"""
 2022 (c) piteren

 NNWrap is an abstract interface of object that wraps neural network (NN) and adds some features.

    NNWrap:
    - Builds FWD and OPT graph. FWD graph is built with nngraph which may be
      an object or callable. FWD graph should be built from inputs to loss.
    - Manages one NNWrap folder (subfolder of SAVE_TOPDIR named with NNWrap name)
      for all NNWrap data (logs, params, checkpoints). NNWrap supports
      serialization into this folder.
    - Extends ParaSave, manages all init parameters. Properly resolves parameters
      using all possible sources, saves and loads them from NNWrap folder.
    - Parameters are kept in self as a Subscriptable to be easily accessed.
    - Properly resolves and holds name of object, adds stamp if needed.
    - Supports / creates logger.
    - May be read only (prevents save over).
    - May be called (with __call__) <- runs NN FWD with given data
    - May be called BWD with backward() <- runs gradient backprop for given data
    - Supports hpmser mode.
    - Manages seed and guarantees reproducibility.
    - Manages GPU / CPU devices used by NN.
    - Adds TensorBoard support.
    - Defines / implements save / load / copy of whole NNWrap (ParaSave + NN checkpoint).
    - Defines interface of baseline training & testing with data loaded to Batcher.
    - Defines / implements GX.
    - Adds some sanity checks.
"""

from abc import ABC, abstractmethod
from typing import Optional, Callable, Union, Tuple, Dict

from pypaq.comoneural.batcher import Batcher
from pypaq.lipytools.little_methods import get_params, get_func_dna
from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.pylogger import get_pylogger, get_hi_child
from pypaq.pms.parasave import ParaSave
from pypaq.torchness.tbwr import TBwr


class NNWrapException(Exception):
    pass


class NNWrap(ParaSave, ABC):

    # restricted keys for fwd_func DNA and return DNA (if they appear in kwargs, should be named exactly like below)
    SPEC_KEYS: set = {
        'loss',                         # loss
        'acc',                          # accuracy
        'f1'}                           # F1

    # defaults (state), may be overridden with kwargs or nngraph (__init__?) attributes
    INIT_DEFAULTS: dict = {
        'seed':             123,        # seed
        'devices':          -1,         # :DevicesParam (check pypaq.mpython.devices)
            # training
        'batch_size':       64,         # training batch size
        'n_batches':        1000,       # default length of training
        'train_batch_IX':   0,          # default (starting) batch index (counter)
            # LR management
        'baseLR':           3e-4,
        'warm_up':          None,
        'ann_base':         None,
        'ann_step':         1.0,
        'n_wup_off':        2.0,
            # gradients clipping parameters
        'clip_value':       None,
        'avt_SVal':         0.1,
        'avt_window':       100,
        'avt_max_upd':      1.5,
        'do_clip':          False,
            # other
        'hpmser_mode':      False,      # it will set model to be read_only and quiet when running with hpmser
        'read_only':        False,      # sets model to be read only - wont save anything (wont even create self.nnwrap_dir)
        'do_TB':            True}       # runs TensorBard, saves in self.nnwrap_dir

    SAVE_TOPDIR = '_models'

    def __init__(
            self,
            nngraph: Optional[Union[Callable, type]]=   None,   # function (Callable) or type (of object) that defines NN
            name: Optional[str]=                        None,
            name_timestamp=                             False,  # adds timestamp to the name
            save_topdir: Optional[str]=                 None,
            save_fn_pfx: Optional[str]=                 None,
            logger=                                     None,
            loglevel=                                   20,
            **kwargs):

        if not save_topdir: save_topdir = self.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = self.SAVE_FN_PFX

        if not name and not nngraph:
            raise NNWrapException ('NNWrap ERROR: name OR nngraph must be given!')

        self.nngraph = nngraph # INFO: here we save TYPE

        self.name = self._generate_name(
            given_name= name,
            timestamp=  name_timestamp)

        # some early overrides

        if kwargs.get('hpmser_mode', False):
            loglevel = 50
            kwargs['read_only'] = True

        if kwargs.get('read_only', False):
            kwargs['do_TB'] = False

        _read_only = kwargs.get('read_only', False)

        self.nnwrap_dir = f'{save_topdir}/{self.name}'

        if not logger:
            logger = get_pylogger(
                name=       self.name,
                add_stamp=  False,
                folder=     None if _read_only else self.nnwrap_dir,
                level=      loglevel)
        self._nwwlog = logger
        nng_info = self.nngraph.__name__ if self.nngraph else 'nngraph NOT GIVEN (will try to load from saved)'
        self._nwwlog.info(f'*** NNWrap *** name: {self.name} initializes for nngraph: {nng_info}')
        self._nwwlog.info(f'> NNWrap dir: {self.nnwrap_dir}{" <- read only mode!" if _read_only else ""}')

        self._dna = dict(
            nngraph=    self.nngraph,
            model_dir=  self.nnwrap_dir)
        self._manage_dna(
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx,
            **kwargs)

        ParaSave.__init__(
            self,
            lock_managed_params=    True,
            logger=                 get_hi_child(self._nwwlog),
            **self._dna)
        self.check_params_sim(params= list(self.SPEC_KEYS) + list(self.INIT_DEFAULTS.keys())) # safety check

        self._manage_devices()

        self._set_seed()

        self._build_graph()

        self._TBwr = TBwr(logdir=self.nnwrap_dir)  # TensorBoard writer

        self._batcher = None

        self._nwwlog.debug(str(self))
        self._nwwlog.info(f'NNWrap init finished!')

    # ******************************************************************************************* NNWrap init submethods

    # generates NNWrap name
    @abstractmethod
    def _generate_name(
            self,
            given_name: Optional[str],
            timestamp: bool) -> str:
        pass

    # manages dna, reports
    def _manage_dna(
            self,
            save_topdir: str,
            save_fn_pfx: str,
            **kwargs) -> None:

        # load dna from folder
        dna_saved = self.load_dna(
            name=           self.name,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx)

        # in case 'nngraph' was not given with init, try to get it from saved
        if not self.nngraph:
            self.nngraph = dna_saved.get('nngraph', None)

        if not self.nngraph:
            msg = 'nngraph was not given and has not been found in saved, cannot continue!'
            self._nwwlog.error(msg)
            raise NNWrapException(msg)

        # get defaults of given nngraph (object.__init__ or callable)
        nngraph_func = self.nngraph.__init__ if type(self.nngraph) is object else self.nngraph
        nngraph_func_params = get_params(nngraph_func)
        nngraph_func_params_defaults = nngraph_func_params['with_defaults']   # get init params defaults
        if 'logger' in nngraph_func_params_defaults: nngraph_func_params_defaults.pop('logger')

        # update in proper order
        self._dna.update(self.INIT_DEFAULTS)
        self._dna.update(nngraph_func_params_defaults)
        self._dna.update(dna_saved)
        self._dna.update(kwargs)  # update with kwargs given NOW by user
        self._dna.update({
            'name':         self.name,
            'save_topdir':  save_topdir,
            'save_fn_pfx':  save_fn_pfx})

        dna_with_nwwlog = {}
        dna_with_nwwlog.update(self._dna)
        dna_with_nwwlog['logger'] = get_hi_child(
            logger= self._nwwlog,
            name=   f'{self.name}_sublogger'),
        self._dna_nngraph = get_func_dna(nngraph_func, dna_with_nwwlog)

        not_used_kwargs = {}
        for k in kwargs:
            if k not in self._dna_nngraph:
                not_used_kwargs[k] = kwargs[k]

        self._nwwlog.debug(f'> {self.name} DNA sources:')
        self._nwwlog.debug(f'>> class INIT_DEFAULTS:        {self.INIT_DEFAULTS}')
        self._nwwlog.debug(f'>> nngraph defaults:           {nngraph_func_params_defaults}')
        self._nwwlog.debug(f'>> DNA saved:                  {dna_saved}')
        self._nwwlog.debug(f'>> given kwargs:               {kwargs}')
        self._nwwlog.debug(f'> resolved DNA:')
        self._nwwlog.debug(f'nngraph complete DNA:          {self._dna_nngraph}')
        self._nwwlog.debug(f'>> kwargs not used by nngraph: {not_used_kwargs}')
        self._nwwlog.debug(f'{self.name} complete DNA:      {self._dna}')

    # manages NNWrap CUDA / CPU devices
    @abstractmethod
    def _manage_devices(self) -> None: pass

    # sets NNWrap seed in all possible areas
    @abstractmethod
    def _set_seed(self) -> None: pass

    # builds NNWrap graph
    @abstractmethod
    def _build_graph(self) -> None: pass

    @abstractmethod
    def __call__(self, *args, **kwargs) -> dict: pass

    @abstractmethod
    def backward(self, *args, **kwargs) -> dict: pass

    # *********************************************************************************************** load / save / copy

    # (re)loads model checkpoint
    @abstractmethod
    def load_ckpt(self) -> None: pass

    # saves model checkpoint
    @abstractmethod
    def save_ckpt(self) -> None: pass

    # saves NNWrap (ParaSave DNA and model checkpoint)
    def save(self):
        if self['read_only']: raise NNWrapException('read only NNWrap cannot be saved!')
        self.save_dna()
        self.save_ckpt()
        self._nwwlog.info(f'NNWrap {self.name} saved')

    # copies just model checkpoint
    @classmethod
    @abstractmethod
    def copy_checkpoint(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None) -> None: pass

    # copies full NNWrap folder (DNA & checkpoints)
    @classmethod
    def copy_saved(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None,
            save_fn_pfx: Optional[str]=     None):

        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        if save_topdir_trg is None: save_topdir_trg = save_topdir_src

        cls.copy_saved_dna(
            name_src=           name_src,
            name_trg=           name_trg,
            save_topdir_src=    save_topdir_src,
            save_topdir_trg=    save_topdir_trg,
            save_fn_pfx=        save_fn_pfx)

        cls.copy_checkpoint(
            name_src=           name_src,
            name_trg=           name_trg,
            save_topdir_src=    save_topdir_src,
            save_topdir_trg=    save_topdir_trg)

    # *************************************************************************************************************** GX

    # GX for two NNWrap checkpoints
    @classmethod
    def gx_ckpt(
            cls,
            name_A: str,                            # name parent A
            name_B: str,                            # name parent B
            name_child: str,                        # name child
            save_topdir_A: Optional[str]=       None,
            save_topdir_B: Optional[str]=       None,
            save_topdir_child: Optional[str]=   None,
            ratio: float=                       0.5,
            noise: float=                       0.03) -> None: pass

    # performs GX on saved NNWrap objects, without even building child objects
    @classmethod
    def gx_saved(
            cls,
            name_parent_main: str,
            name_parent_scnd: Optional[str],    # if not given makes GX only with main parent
            name_child: str,
            save_topdir_parent_main: Optional[str]= None,
            save_topdir_parent_scnd: Optional[str]= None,
            save_topdir_child: Optional[str]=       None,
            save_fn_pfx: Optional[str]=             None,
            do_gx_ckpt=                             True,
            ratio: float=                           0.5,
            noise: float=                           0.03
    ) -> None:

        if not save_topdir_parent_main: save_topdir_parent_main = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        cls.gx_saved_dna(
            name_parent_main=           name_parent_main,
            name_parent_scnd=           name_parent_scnd,
            name_child=                 name_child,
            save_topdir_parent_main=    save_topdir_parent_main,
            save_topdir_parent_scnd=    save_topdir_parent_scnd,
            save_topdir_child=          save_topdir_child,
            save_fn_pfx=                save_fn_pfx)

        if do_gx_ckpt:
            cls.gx_ckpt(
                name_A=             name_parent_main,
                name_B=             name_parent_scnd or name_parent_main,
                name_child=         name_child,
                save_topdir_A=      save_topdir_parent_main,
                save_topdir_B=      save_topdir_parent_scnd,
                save_topdir_child=  save_topdir_child,
                ratio=              ratio,
                noise=              noise)
        else:
            cls.copy_checkpoint(
                name_src=           name_parent_main,
                name_trg=           name_child,
                save_topdir_src=    save_topdir_parent_main,
                save_topdir_trg=    save_topdir_child)

    # ***************************************************************************************************** train / test

    # loads data to Batcher
    def load_data(
            self,
            data_TR: Dict,
            data_VL: Optional[Dict]=    None,
            data_TS: Optional[Dict]=    None):

        self._batcher = Batcher(
            data_TR=        data_TR,
            data_VL=        data_VL,
            data_TS=        data_TS,
            batch_size=     self['batch_size'],
            batching_type=  'random_cov',
            logger=         get_hi_child(self._nwwlog, 'Batcher'))

     # trains model, returns optional test score
    @abstractmethod
    def run_train(
            self,
            data=                       None,
            n_batches: Optional[int]=   None,
            test_freq=                  100,    # number of batches between tests, model SHOULD BE tested while training
            mov_avg_factor=             0.1,
            save_max=                   True,   # allows to save model while training (after max test)
            use_F1=                     True,   # uses F1 as a train/test score (not acc)
            **kwargs) -> Optional[float]: pass

    # tests model, returns: optional accuracy, optional F1, loss (average)
    @abstractmethod
    def run_test(self, **kwargs) -> Tuple[Optional[float], Optional[float], float]: pass

    # updates model baseLR
    @abstractmethod
    def update_baseLR(self, lr: float) -> None: pass

    @property
    def tbwr(self):
        return self._TBwr

    # logs value to TB
    def log_TB(
            self,
            value,
            tag: str,
            step: int) -> None:
        if self['do_TB']: self._TBwr.add(value=value, tag=tag, step=step)
        else: self._nwwlog.warning(f'NNWrap {self.name} cannot log TensorBoard since do_TB flag is False!')

    # returns nice string about self
    @abstractmethod
    def __str__(self) -> str: pass


from typing import Optional

import numpy as np
import torch

from pypaq.lipytools.pylogger import get_pylogger


# applies warm-up at the beginning (for warm_up steps) and annealing after some steps (after warm_up * n_wup_off), should be called every step
class ScaledLR(torch.optim.lr_scheduler._LRScheduler):

    def __init__(
            self,
            optimizer,
            starting_step: int=         0,      # step to start with
            warm_up: Optional[int]=     1000,   # number of steps for linear warm-up, for None or 0 warm-up is turned off
            ann_base: Optional[float]=  0.999,  # annealing base, None or 1.0 turns off annealing
            ann_step: float=            1.0,    # annealing step, higher value speeds up annealing
            n_wup_off: float=           2.0,    # number of warm-up durations to start annealing
            last_epoch=                 -1,
            logger=                     None):

        if not logger: logger = get_pylogger()
        self._log = logger

        self._step = starting_step
        self.warm_up = warm_up or 0
        self.ann_base = ann_base
        self.ann_step = ann_step
        self.n_wup_off = n_wup_off

        super(ScaledLR, self).__init__(optimizer, last_epoch, verbose=self._log.getEffectiveLevel()<20)

    # updates LR of 0 group
    def update_base_lr0(self, lr: float):
        self.base_lrs[0] = lr

    def get_lr(self):

        lrs = np.array(self.base_lrs) # self.base_lrs keeps [baseLR] of groups
        if self.warm_up:
            wm_ratio = min(self._step, self.warm_up) / self.warm_up
            lrs *= wm_ratio
            self._log.debug(f'applied warmUp ({self.warm_up}) to lR')

        if self.ann_base is not None and self.ann_base != 1.0:
            steps_offs = max(0, self._step - int(self.warm_up * self.n_wup_off))
            lrs *= self.ann_base ** (steps_offs * self.ann_step)
            self._log.debug(f'applied annealing to lR ({self.ann_base:.5f},{self.ann_step:.5f})')

        self._log.debug(f'ScaledLR scheduler step: {self._step} lrs: {lrs.tolist()}')
        self._step += 1
        return lrs.tolist()

    def _get_closed_form_lr(self):
        return self.get_lr()



from typing import Optional, Callable

from pypaq.R4C.policy_gradients.base.tf_based.pg_TF_actor import PG_TFActor
from pypaq.R4C.policy_gradients.a2c.tf_based.a2c_TF_graph import a2c_graph



class A2C_TFActor(PG_TFActor):

    def __init__(
            self,
            name: str=                      'A2C_TFActor',
            nngraph: Optional[Callable]=    a2c_graph,
            **kwargs):
        PG_TFActor.__init__(
            self,
            name=       name,
            nngraph=    nngraph,
            **kwargs)

    # overrides PG_TFActor with more log_TB
    def update_with_experience(
            self,
            observations,
            actions,
            dreturns,
            inspect=    False) -> dict:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw.backward(
            feed_dict=  {
                self.nnw['observation_PH']:  obs_vecs,
                self.nnw['action_PH']:       actions,
                self.nnw['return_PH']:       dreturns},
            fetch=      ['optimizer','probs','loss','loss_actor','loss_critic','gg_norm','gg_avt_norm','actor_ce_mean','zeroes'])
        out.pop('optimizer')
        return out


"""

 2022 (c) piteren

    DQNTFActor - DQNActor implemented with NEModel (TF)

"""

import numpy as np
from typing import List, Optional, Callable

from pypaq.R4C.qlearning.dqn.dqn_actor import DQN_Actor
from pypaq.R4C.qlearning.dqn.tf_based.dqn_TF_graph import dqn_graph
from pypaq.neuralmess.nemodel import NEModel


# DQN (TF NN) based QLearningActor
class DQN_TFActor(DQN_Actor):

    def __init__(
            self,
            name: str=                      'DQN_TFActor',
            nngraph: Optional[Callable]=    dqn_graph,
            **kwargs):
        DQN_Actor.__init__(
            self,
            name=       name,
            nnwrap=     NEModel,
            nngraph=    nngraph,
            **kwargs)

    def _get_QVs(self, observation: object) -> np.ndarray:
        obs_vec = self._get_observation_vec(observation)
        out = self.nnw(
            feed_dict=  {self.nnw['observations_PH']: [obs_vec]},
            fetch=      ['output'])
        return out['output'][0] # reduce dim

    # optimized with single call to session with a batch of observations
    def get_QVs_batch(self, observations: List[object]) -> np.ndarray:
        obs_vecs = np.array([self._get_observation_vec(o) for o in observations])
        out = self.nnw(
            feed_dict=  {self.nnw['observations_PH']: obs_vecs},
            fetch=      ['output'])
        return out['output']

    # optimized with single call to session with a batch of data
    def update_with_experience(
            self,
            observations: List[object],
            actions: List[int],
            new_qvs: List[float],
            inspect=    False) -> dict:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw.backward(
            feed_dict=  {
                self.nnw['observations_PH']: obs_vecs,
                self.nnw['enum_actions_PH']: np.array(list(enumerate(actions))),
                self.nnw['gold_QV_PH']:      np.array(new_qvs)},
            fetch=      ['optimizer','loss','gg_norm','gg_avt_norm'])
        out.pop('optimizer')
        return out


"""

 2022 (c) piteren

    QLearningTrainer for QLearningActor acting on FiniteActionsRLEnvy.

"""

import numpy as np

from pypaq.R4C.helpers import extract_from_batch
from pypaq.R4C.trainer import FATrainer
from pypaq.R4C.qlearning.ql_actor import QLearningActor


# Q-Learning Trainer
class QLearningTrainer(FATrainer):

    def __init__(
            self,
            actor: QLearningActor,
            gamma: float,       # QLearning gamma (discount)
            **kwargs):

        FATrainer.__init__(self, actor=actor, **kwargs)
        self.actor = actor  # INFO: just type "upgrade" for pycharm editor
        self.gamma = gamma

        self._rlog.info(f'*** QLearningTrainer *** initialized')
        self._rlog.info(f'> gamma: {self.gamma}')

    # updates QLearningActor policy with batch of random data from memory
    def _update_actor(self, inspect=False) -> dict:

        batch = self.memory.sample(self.batch_size)

        observations =      extract_from_batch(batch, 'observation')
        actions =           extract_from_batch(batch, 'action')
        rewards =           extract_from_batch(batch, 'reward')
        next_observations = extract_from_batch(batch, 'next_observation')
        terminals =         extract_from_batch(batch, 'terminal')

        no_qvs = self.actor.get_QVs_batch(next_observations)
        no_qvs_terminal = np.zeros(self.envy.num_actions())

        for ix,t in enumerate(terminals):
            if t: no_qvs[ix] = no_qvs_terminal

        new_qvs = [(r + self.gamma * max(no_qvs)) for r, no_qvs in zip(rewards, no_qvs)]

        return self.actor.update_with_experience(
            observations=   observations,
            actions=        actions,
            new_qvs=        new_qvs,
            inspect=        inspect)


"""

 2022 (c) piteren

    RL Actor

"""

from abc import abstractmethod, ABC
import numpy as np
from typing import Optional

from pypaq.lipytools.little_methods import stamp
from pypaq.lipytools.pylogger import get_pylogger
from pypaq.R4C.envy import RLEnvy
from pypaq.R4C.helpers import RLException


# just Actor
class Actor(ABC):

    # returns Actor action based on observation according to Actor policy
    @abstractmethod
    def get_policy_action(self, observation: object) -> object: pass

# cooperates with RLTrainer, prepares obs_vec, updates, saves
class TrainableActor(Actor, ABC):

    def __init__(
            self,
            envy: RLEnvy,
            name: str=              'TrainableActor',
            name_timestamp: bool=   True,
            logger: Optional=       None,
            loglevel: int=          20,
            **kwargs):

        if name_timestamp: name += f'_{stamp()}'
        self.name = name
        self._envy = envy

        self._rlog = logger or get_pylogger(level=loglevel)
        self._rlog.info(f'*** TrainableActor *** initialized')
        self._rlog.info(f'> name:              {self.name}')
        self._rlog.info(f'> Envy:              {self._envy.__class__.__name__}')
        self._rlog.info(f'> observation width: {self._get_observation_vec(self._envy.get_observation()).shape[-1]}')
        self._rlog.info(f'> not used kwargs:   {kwargs}')

    # prepares numpy vector from observation, first tries to get from RLEnvy
    def _get_observation_vec(self, observation: object) -> np.ndarray:
        try: return self._envy.prep_observation_vec(observation)
        except RLException: raise RLException ('TrainableActor not implemented _get_observation_vec()')

    # add sampling (from probability?) option which may be helpful for training
    @abstractmethod
    def get_policy_action(self, observation:object, sampled=False) -> object: pass

    # updates self with (batch of) experience data (given with kwargs), returns dict with Actor "metrics" - loss etc.
    @abstractmethod
    def update_with_experience(self, **kwargs) -> dict: pass

    @abstractmethod
    # returns Actor TOP save directory
    def _get_save_topdir(self) -> str: pass

    # returns Actor save directory
    def get_save_dir(self) -> str:
        return f'{self._get_save_topdir()}/{self.name}'

    # saves (self) Actor state
    @abstractmethod
    def save(self): pass

    # returns some info about Actor
    @abstractmethod
    def __str__(self) -> str: pass


from abc import ABC

from typing import Optional, Callable

from pypaq.R4C.policy_gradients.base.tf_based.pg_TF_actor import PG_TFActor
from pypaq.R4C.policy_gradients.actor_critic_shared.tf_based.ac_shared_TF_graph import acs_graph


class ACShared_TFActor(PG_TFActor, ABC):

    def __init__(
            self,
            name: str=                      'ACShared_TFActor',
            nngraph: Optional[Callable]=    acs_graph,
            **kwargs):
        PG_TFActor.__init__(
            self,
            name=       name,
            nngraph=    nngraph,
            **kwargs)

    def update_with_experience(
            self,
            observations,
            actions,
            dreturns,
            inspect=    False) -> dict:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw.backward(
            feed_dict=  {
                self.nnw['observation_PH']:  obs_vecs,
                self.nnw['action_PH']:       actions,
                self.nnw['qv_label_PH']:     dreturns},
            fetch=      ['optimizer','loss','loss_actor','loss_critic','gg_norm','gg_avt_norm','amax_prob','amin_prob'])
        out.pop('optimizer')
        return out